from comfit.tool.tool_print_in_color import tool_print_in_color

def tool_configure_axis(dim, name, xlim=None, xmin=None, xmax=None, xRes=None, dx=None):
        """
        Configure a single axis (x, y, or z) based on the provided parameters.
        Parameters:
            dim (int): Dimension of the system (1, 2, or 3).
            name (str): Name of the axis (e.g., 'x', 'y', 'z') for error messages.
            xlim (list): A 2-element list [min, max] for the axis limits.
            xmin (float): Minimum value of the axis.
            xmax (float): Maximum value of the axis.
            xRes (int): Resolution (number of points) along the axis.
            dx (float): Step size (spacing between points) along the axis.
        Returns:
            xlim, xmin, xmax, xRes, dx
        """

        #Hierarcy of values: lim > min_val > max_val > res > delta

        # Find the axis limits
        xlim_YUP_provided = xlim is not None
        xlim_NOT_provided = xlim is None
        
        xmin_YUP_provided = xmin is not None
        xmin_NOT_provided = xmin is None

        xmax_YUP_provided = xmax is not None
        xmax_NOT_provided = xmax is None

        xRes_YUP_provided = xRes is not None
        xRes_NOT_provided = xRes is None

        dx_YUP_provided = dx is not None
        dx_NOT_provided = dx is None

        if xlim_NOT_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_NOT_provided and dx_NOT_provided:
            # E.g. no values provided

            xmin = 0
            xmax = 101 
            xRes = 101
            dx = 1.0

            if (name == 'y' and dim < 2) or (name == 'z' and dim < 3):
                xmax = 1
                xRes = 1
                
            xlim = [xmin, xmax]
            
        elif xlim_NOT_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_NOT_provided and dx_YUP_provided:
            # E.g. dx = 0.1

            xmin = 0
            xmax = 101
            xRes = round((xmax - xmin) / dx)
            xlim = [xmin, xmax]
        
        elif xlim_NOT_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_YUP_provided and dx_NOT_provided:
            # E.g. xRes = 56

            xmin = 0
            xmax = xRes
            dx = (xmax - xmin) / xRes
            xlim = [xmin, xmax]
        
        elif xlim_NOT_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_YUP_provided and dx_YUP_provided:
            # E.g. xRes = 56, dx = 0.1
            
            xmin = 0
            xmax = xmin + xRes * dx
            xlim = [xmin, xmax]
        
        elif xlim_NOT_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_NOT_provided and dx_NOT_provided:
            # E.g. xmax = 56
            
            xmin = 0 if xmax > 0 else xmax-101
            xRes = 101
            dx = (xmax - xmin) / xRes
            xlim = [xmin, xmax]
        
        elif xlim_NOT_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_NOT_provided and dx_YUP_provided:
            # E.g. xmax = 21, dx = 0.1

            xmin = 0 if xmax > 0 else xmax-101
            xRes = round((xmax - xmin) / dx)
            xlim = [xmin, xmax]
        
        elif xlim_NOT_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_YUP_provided and dx_NOT_provided:
            # E.g. xmax = 21, xRes = 56

            xmin = 0 if xmax > 0 else xmax-xRes
            dx = (xmax - xmin) / xRes
            xlim = [xmin, xmax]

        elif xlim_NOT_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_YUP_provided and dx_YUP_provided:
            # E.g. xmax = 21, xRes = 56, dx = 0.1

            xmin = xmax - xRes * dx
            xlim = [xmin, xmax]
        
        elif xlim_NOT_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_NOT_provided and dx_NOT_provided:
            # E.g. xmin = -5

            xmax = xmin+101
            xRes = 101
            dx = 1.0
            xlim = [xmin, xmax]
        
        elif xlim_NOT_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_NOT_provided and dx_YUP_provided:
            # E.g. xmin = -5, dx = 0.1

            xRes = 101
            xmax = xmin + xRes * dx
            xlim = [xmin, xmax]
        
        elif xlim_NOT_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_YUP_provided and dx_NOT_provided:
            # E.g. xmin = -5, xRes = 56

            dx = 1.0
            xmax = xmin + xRes * dx
            xlim = [xmin, xmax]
        
        elif xlim_NOT_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_YUP_provided and dx_YUP_provided:
            # E.g. xmin = -5, xRes = 56, dx = 0.1

            xmax = xmin + xRes * dx
            xlim = [xmin, xmax]
        
        elif xlim_NOT_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_NOT_provided and dx_NOT_provided:
            # E.g. xmin = -5, xmax = 21

            xRes = xmax - xmin
            dx = 1.0
            xlim = [xmin, xmax]
        
        elif xlim_NOT_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_NOT_provided and dx_YUP_provided:
            # E.g. xmin = -5, xmax = 21, dx = 0.1

            xRes = round((xmax - xmin) / dx)
            dx = (xmax - xmin) / xRes 
            xlim = [xmin, xmax]

        elif xlim_NOT_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_YUP_provided and dx_NOT_provided:
            # E.g. xmin = -5, xmax = 21, xRes = 56

            dx = (xmax - xmin) / xRes
            xlim = [xmin, xmax]
        
        elif xlim_NOT_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_YUP_provided and dx_YUP_provided:
            # E.g. xmin = -5, xmax = 21, xRes = 56, dx = 0.1

            tool_print_in_color(f'Warning: Providing {name}min, {name}max, and {name}Res trumps providing d{name}.', 'yellow')
            dx = (xmax - xmin) / xRes
            xlim = [xmin, xmax]
        
        elif xlim_YUP_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_NOT_provided and dx_NOT_provided:
            # E.g. xlim = [-5, 21]

            xmin = xlim[0]
            xmax = xlim[1]
            xRes = 101
            dx = (xmax - xmin) / xRes
        
        elif xlim_YUP_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_NOT_provided and dx_YUP_provided:
            # E.g. xlim = [-5, 21], dx = 0.1

            xmin = xlim[0]
            xmax = xlim[1]
            xRes = round((xmax - xmin) / dx)
            dx = (xmax - xmin) / xRes
        
        elif xlim_YUP_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_YUP_provided and dx_NOT_provided:
            # E.g. xlim = [-5, 21], xRes = 56

            xmin = xlim[0]
            xmax = xlim[1]
            dx = (xmax - xmin) / xRes
        
        elif xlim_YUP_provided and xmin_NOT_provided and xmax_NOT_provided and xRes_YUP_provided and dx_YUP_provided:
            # E.g. xlim = [-5, 21], xRes = 56, dx = 0.1

            tool_print_in_color(f'Warning: Providing {name}lim and xRes trumps providing d{name}.', 'yellow')
            
            xmin = xlim[0]
            xmax = xlim[1]
            dx = (xmax - xmin) / xRes
        
        elif xlim_YUP_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_NOT_provided and dx_NOT_provided:
            # E.g. xlim = [-5, 21], xmax = 21

            tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}max.', 'yellow')

            xmin = xlim[0]
            xmax = xlim[1]
            xRes = 101
            dx = (xmax - xmin) / xRes

        elif xlim_YUP_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_NOT_provided and dx_YUP_provided:
            # E.g. xlim = [-5, 21], xmax = 21, dx = 0.1

            tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}max.', 'yellow')

            xmin = xlim[0]
            xmax = xlim[1]
            xRes = round((xmax - xmin) / dx)
            dx = (xmax - xmin) / xRes
        
        elif xlim_YUP_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_YUP_provided and dx_NOT_provided:
            # E.g. xlim = [-5, 21], xmax = 21, xRes = 56

            tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}max.', 'yellow')

            xmin = xlim[0]
            xmax = xlim[1]
            dx = (xmax - xmin) / xRes
        
        elif xlim_YUP_provided and xmin_NOT_provided and xmax_YUP_provided and xRes_YUP_provided and dx_YUP_provided:
            # E.g. xlim = [-5, 21], xmax = 21, xRes = 56, dx = 0.1

            tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}max.', 'yellow')
            tool_print_in_color(f'Warning: Providing {name}Res and {name}lim trumps providing d{name}.', 'yellow')

            xmin = xlim[0]
            xmax = xlim[1]
            dx = (xmax - xmin) / xRes

        elif xlim_YUP_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_NOT_provided and dx_NOT_provided:
            # E.g. xlim = [-5, 21], xmin = -5

            tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min.', 'yellow')

            xmin = xlim[0]
            xmax = xlim[1]
            xRes = 101
            dx = (xmax - xmin) / xRes

        elif xlim_YUP_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_NOT_provided and dx_YUP_provided:
            # E.g. xlim = [-5, 21], xmin = -5, dx = 0.1

            tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min.', 'yellow')

            xmin = xlim[0]
            xmax = xlim[1]
            xRes = round((xmax - xmin) / dx)
            dx = (xmax - xmin) / xRes

        elif xlim_YUP_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_YUP_provided and dx_NOT_provided:
            # E.g. xlim = [-5, 21], xmin = -5, xRes = 56

            tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min.', 'yellow')

            xmin = xlim[0]
            xmax = xlim[1]
            dx = (xmax - xmin) / xRes
        
        elif xlim_YUP_provided and xmin_YUP_provided and xmax_NOT_provided and xRes_YUP_provided and dx_YUP_provided:
            # E.g. xlim = [-5, 21], xmin = -5, xRes = 56, dx = 0.1

            tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min.', 'yellow')
            tool_print_in_color(f'Warning: Providing {name}Res and {name}lim trumps providing d{name}.', 'yellow')

            xmin = xlim[0]
            xmax = xlim[1]
            dx = (xmax - xmin) / xRes
        
        elif xlim_YUP_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_NOT_provided and dx_NOT_provided:
            # E.g. xlim = [-5, 21], xmin = -5, xmax = 21

            tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min and {name}max.', 'yellow')

            xmin = xlim[0]
            xmax = xlim[1]
            xRes = 101
            dx = (xmax - xmin) / xRes
        
        elif xlim_YUP_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_NOT_provided and dx_YUP_provided:
            # E.g. xlim = [-5, 21], xmin = -5, xmax = 21, dx = 0.1

            tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min and {name}max.', 'yellow')

            xmin = xlim[0]
            xmax = xlim[1]
            xRes = round((xmax - xmin) / dx)
            dx = (xmax - xmin) / xRes
        
        elif xlim_YUP_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_YUP_provided and dx_NOT_provided:
            # E.g. xlim = [-5, 21], xmin = -5, xmax = 21, xRes = 56

            tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min and {name}max.', 'yellow')

            xmin = xlim[0]
            xmax = xlim[1]
            dx = (xmax - xmin) / xRes
        
        elif xlim_YUP_provided and xmin_YUP_provided and xmax_YUP_provided and xRes_YUP_provided and dx_YUP_provided:
            # E.g. xlim = [-5, 21], xmin = -5, xmax = 21, xRes = 56, dx = 0.1

            tool_print_in_color(f'Warning: Providing {name}lim trumps providing {name}min and {name}max.', 'yellow')
            tool_print_in_color(f'Warning: Providing {name}Res and {name}lim trumps providing d{name}.', 'yellow')

            xmin = xlim[0]
            xmax = xlim[1]
            dx = (xmax - xmin) / xRes
        
        return xlim, xmin, xmax, xRes, dx


import numpy as np
import scipy as sp

class BaseSystemGet:
    """ Get methods for the base system class"""
    def get_sym(self, tensor: np.ndarray, i: int, j: int) -> np.ndarray:
        """Gets the i,j component of a symmetric tensor saved in an array structure.

        Args: 
            tensor (numpy.ndarray): The symmetric tensor.
            i: The row index. (int)
            j: The column index. (int)
        
        Returns:
            The i,j component of the tensor. (numpy.ndarray) 
        """

        if self.dim == 2:
            if i == 0:
                return tensor[0] if j == 0 else tensor[1]
            elif i == 1:
                return tensor[1] if j == 0 else tensor[2]

        elif self.dim == 3:
            if i == 0:
                return tensor[0] if j == 0 else tensor[1] if j == 1 else tensor[2]
            elif i == 1:
                return tensor[1] if j == 0 else tensor[3] if j == 1 else tensor[4]
            elif i == 2:
                return tensor[2] if j == 0 else tensor[4] if j == 1 else tensor[5]

    def get_sym_tl(self, tensor: np.ndarray, i: int, j: int) -> np.ndarray:
        """Gets the i,j component of a symmetric traceless tensor saved in an array structure.

        Args:
            tensor: The symmetric traceless tensor. (numpy.ndarray)
            i: The row index. (int)
            j: The column index. (int)
        
        Returns:
            The i,j component of the tensor. (numpy.ndarray) 
        """
        if self.dim == 2:
            if i == 0:
                return tensor[0] if j == 0 else tensor[1]
            elif i == 1:
                return tensor[1] if j == 0 else -tensor[0]
            
        elif self.dim == 3:
            if i == 0:
                return tensor[0] if j == 0 else tensor[1] if j == 1 else tensor[2]
            elif i == 1:
                return tensor[1] if j == 0 else tensor[3] if j == 1 else tensor[4]
            elif i == 2:
                return tensor[2] if j == 0 else tensor[4] if j == 1 else -(tensor[0] + tensor[3])

    def get_anti_sym(self, omega: np.ndarray, i: int, j: int) -> np.ndarray:
        """Gets the i,j component of an anti-symmetric tensor saved in an array structure.

        Args:
            omega: The anti-symmetric tensor. (numpy.ndarray)
            i: The row index. (int)
            j: The column index. (int)
        
        Returns:
            The i,j component of the tensor. (numpy.ndarray) 
        """
        # TODO: I don't like that the input vector is a scalar field in 2 dimensions. (Vidar 11.03.24)
        if self.dim == 2:
            if i==j:
                return 0
            return (-1)**i *omega
        elif self.dim ==3:
            if i ==j:
                return 0
            else:
                return np.sign(j-i)*omega[i+j-1]



import numpy as np
import plotly.graph_objects as go
import plotly.figure_factory as ff
from comfit.tool.tool_complete_field import tool_complete_field
from comfit.tool.tool_set_plot_axis_properties_plotly import tool_set_plot_axis_properties_plotly
from comfit.tool.tool_add_spacing_2D import tool_add_spacing_2D

from comfit.plot.plot_field_plotly import plot_field_plotly


def plot_field_velocity_and_director_plotly(self, field, velocity, director, **kwargs):
    """Plot the fields, velocity, and director field in 2 dimensions using Plotly.

    Args:
        field (ndarray): The field to be plotted.
        velocity (ndarray): The velocity to be plotted.
        director (ndarray): The director to be plotted.
        **kwargs: Keyword arguments for the plot.

    Returns:
        The plotly figure (go.Figure).
    """
    if field.dtype == bool:
        field = field.astype(float)

    # Check if the vector field is complex
    if np.iscomplexobj(field):
        print(
                "\033[91mWarning: the provided field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
        print('Max imaginary part: ', np.max(np.imag(field)))
    field = np.real(field)

    # Check if an axis object is provided
    fig = kwargs.get('fig', go.Figure())

    # Extend the field if not a complete array is given
    field = tool_complete_field(self, field)

    kwargs['colormap'] = kwargs.get('colormap', 'Picnic')
    kwargs['vlim_symmetric'] = kwargs.get('vlim_symmetric', True)

    fig_tmp = plot_field_plotly(self, field, **kwargs)
    fig.add_trace(fig_tmp.data[0].update(showlegend=False))

    # Plot the director field using quiver
    X, Y = np.meshgrid(self.x, self.y, indexing='ij')

    spacing = kwargs.get('spacing', 5)

    X, Y, U, V = tool_add_spacing_2D(X,Y,director[0],director[1],spacing)

    u = U.flatten()
    v = V.flatten()

    magnitude = np.sqrt(u**2 + v**2)
    magnitude_max = max(np.max(magnitude),1e-12)
    magnitude_normalized = magnitude/magnitude_max

    angle = np.arctan2(v, u)
    direction = np.array([np.cos(angle), np.sin(angle)]).T
    
    fig.add_trace(go.Scatter(
        x=X.flatten()/self.a0,
        y=Y.flatten()/self.a0,
        mode='markers',
        marker=dict(symbol='line-ew', 
            angle=90-angle.flatten()*180/np.pi, 
            size=2*spacing*magnitude_normalized.flatten(), 
            sizemode='diameter',
            color=magnitude.flatten(), 
            cmin=0,
            cmax=magnitude_max,
            line=dict(color='black')
            ),
            hovertemplate='<b>x:</b> %{x:.2f}a0<br>' +
                '<b>y:</b> %{y:.2f}a0<br>' +
                '<b>nx:</b> %{customdata[0]:.2e}<br>' +  
                '<b>ny:</b> %{customdata[1]:.2e}<br>',
            customdata=np.stack((u.flatten(), v.flatten()), axis=-1), showlegend=False  # Adding ux, uy and u as customdata
        )
        )

    fig_tmp = ff.create_streamline(x=self.x/self.a0, 
                                    y=self.y.flatten()/self.a0, 
                                    u=velocity[0].T, 
                                    v=velocity[1].T, 
                                    arrow_scale=1, 
                                    density=1)
        
    for trace in fig_tmp.data:
        trace.showlegend = False

    fig.add_trace(fig_tmp.data[0])

    kwargs['fig'] = fig
    tool_set_plot_axis_properties_plotly(self, **kwargs)
    return fig


import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from skimage.measure import marching_cubes

def plot_surface_matplotlib(self, **kwargs) -> matplotlib.axes.Axes:
    """Plots the surface of the given field.

    Args:
        **kwargs: Keyword arguments for the plot.
    
    Returns:
        The axes containing the plot.
    """
    field = kwargs['field']
    value = kwargs['value']
    ax = kwargs.get('ax', plt.gca())
    alpha = kwargs.get('alpha', 0.5)
    color = kwargs.get('color', 'b')

    verts, faces, _, _ = marching_cubes(field, value)
    x = (self.xmin+verts[:, 0]*self.dx)/self.a0
    y = (self.ymin+verts[:, 1]*self.dy)/self.a0
    z = (self.zmin+verts[:, 2]*self.dz)/self.a0
    ax.plot_trisurf(x, y, faces, z, alpha=alpha, color=color)

    return ax


import numpy as np
import matplotlib.pyplot as plt
import scipy as sp
from typing import Optional
from comfit.tool.tool_complete_field import tool_complete_field
from comfit.tool.tool_colormaps import tool_colormap_angle
from comfit.tool.tool_set_plot_axis_properties_plotly import tool_set_plot_axis_properties_plotly


from skimage.measure import marching_cubes
import plotly.graph_objects as go



def plot_complex_field_in_plane_plotly(
            self,
            complex_field: np.ndarray,
            normal_vector: Optional[np.ndarray] = None,
            position: Optional[np.ndarray] = None,
            **kwargs
        ):
        """Plots the complex field in a plane perpendicular to the given normal vector using

        Args:
            complex_field (array-like): The complex field to be plotted.
            normal_vector (array-like, optional): The normal vector of the plane. Default is [0,1,0].
            position (array-like, optional): The position of the plane. Default is the middle of the system.
            **kwargs: Keyword arguments for the plot, see https://comfitlib.com/Plotting/.

        Returns:
            The axes containing the plot of the complex field. (matplotlib.axes.Axes)
        """

        if self.dim != 3:
            raise Exception("The plot in plane function is only defined for 3D fields.")

        fig = kwargs.get('fig', go.Figure())

        # Extend the field if not a complete array is given
        complex_field = tool_complete_field(self, complex_field)

        # Default values of position and normal vector
        if position is None:
            position = self.rmid

        if normal_vector is None:
            normal_vector = [0,1,0]

        # Calculate the magnitude and phase of the complex field
        rho = np.abs(complex_field)
        theta = np.angle(complex_field)

        reals = np.real(complex_field)
        imags = np.imag(complex_field)

        # Kewyord arguments
        colorbar = kwargs.get('colorbar', True)

        normal_vector = np.array(normal_vector)/np.linalg.norm(normal_vector)
        height_above_plane = (self.x-position[0])*normal_vector[0] + (self.y-position[1])*normal_vector[1] + (self.z-position[2])*normal_vector[2]

        verts, faces, _, _ = marching_cubes(height_above_plane, 0)

        # Calculate the centroids of each triangle
        centroids = np.mean(verts[faces], axis=1)

        # Assuming field is defined on the same grid as height_above_plane
        x, y, z = np.mgrid[0:height_above_plane.shape[0], 0:height_above_plane.shape[1], 0:height_above_plane.shape[2]]

        # Flatten the grid for interpolation
        points = np.c_[x.ravel(), y.ravel(), z.ravel()]
        theta_values = theta.ravel()
        rho_values = rho.ravel()

        # Interpolate field at the vertices positions
        # theta_verts = sp.interpolate.griddata(points, theta_values, centroids, method='nearest')
        # rho_verts = sp.interpolate.griddata(points, rho_values, centroids, method='nearest')

        interpolation_method = kwargs.get('interpolation_method', 'linear') #TODO: This line is inconsistent in naming with
        print("Interpolating points with method: ' ", interpolation_method, "'.")
        print("If this process is slow, consider passing 'interpolation_method='nearest' with the plot_complex_field_in_plane function.") #TODO: this line. (interpolation vs interpolation_method)
        print("That will speed up the process, but the plot may look less smooth.")
        reals_verts = sp.interpolate.griddata(points, reals.ravel(), centroids, method=interpolation_method)
        imags_verts = sp.interpolate.griddata(points, imags.ravel(), centroids, method=interpolation_method)
        print("Interpolation done.")

        theta_verts = np.arctan2(imags_verts, reals_verts)
        rho_verts = np.abs(reals_verts + 1j*imags_verts)

        # Normalize field values for color mapping
        theta_normalized = (theta_verts+np.pi) / (2*np.pi)

        # Map normalized field values to colors
        colormap = tool_colormap_angle()
        colors = colormap(theta_normalized)

        # Blend the colors with white according to rho (normalized)
        colors[:,3] = (rho_verts/np.max(rho_verts)).ravel()

        # Create the mesh object
        x_new = (verts[:, 0] * self.dx + self.xmin) / self.a0
        y_new = (verts[:, 1] * self.dy + self.ymin) / self.a0
        z_new = (verts[:, 2] * self.dz + self.zmin) / self.a0

        mesh = go.Mesh3d(
                    x=x_new, 
                    y=y_new, 
                    z=z_new,
                    i=faces[:, 0], 
                    j=faces[:, 1], 
                    k=faces[:, 2],
                    facecolor=colors,  # Set color for each face
                    showscale=True
                )
        fig.add_trace(mesh)

        # Create a colorbar
        # if colorbar:
        #     mappable = plt.cm.ScalarMappable(cmap=tool_colormap_angle())
        #     mappable.set_array([])
        #     mappable.set_clim(-np.pi, np.pi)
        #     cbar = plt.colorbar(mappable, ax=ax, pad=0.2)
        #     cbar.set_ticks(np.array([-np.pi, -2 * np.pi / 3, -np.pi / 3, 0, np.pi / 3, 2 * np.pi / 3, np.pi]))
        #     cbar.set_ticklabels([r'$-\pi$', r'$-2\pi/3$', r'$-\pi/3$', r'$0$', r'$\pi/3$', r'$2\pi/3$', r'$\pi$'])


        kwargs['plot_is_3D'] = True
        kwargs['fig'] = fig
        tool_set_plot_axis_properties_plotly(self, **kwargs)

        return fig


from .phase_field_crystal_1d_periodic import PhaseFieldCrystal1DPeriodic
from .phase_field_crystal_2d_triangular import PhaseFieldCrystal2DTriangular
from .phase_field_crystal_2d_square import PhaseFieldCrystal2DSquare
from .phase_field_crystal_3d_body_centered_cubic import PhaseFieldCrystal3DBodyCenteredCubic
from .phase_field_crystal_3d_face_centered_cubic import PhaseFieldCrystal3DFaceCenteredCubic
from .phase_field_crystal_3d_simple_cubic import PhaseFieldCrystal3DSimpleCubic


from comfit.phase_field_crystal.phase_field_crystal import PhaseFieldCrystal

import numpy as np
from tqdm import tqdm
from scipy.optimize import fsolve
import scipy as sp
import matplotlib.pyplot as plt
from pprint import pprint

from comfit.tool.tool_print_in_color import tool_print_in_color

class PhaseFieldCrystal2DTriangular(PhaseFieldCrystal):
    def __init__(self, nx, ny, **kwargs):
        """Initializes a phase field crystal system in 2D with a triangular crystal structure.

        Args:
            nx: The number of unit cells in the x direction.
            ny: The number of unit cells in the y direction.

        Returns:
            The system object representing the PhaseFieldCrystal2DTriangular simulation.
        """

        # Type of the system
        self.type = 'PhaseFieldCrystal2DTriangular'
        self.dim = 2
        self.nx = nx
        self.ny = ny

        # Default simulation parameters
        self.micro_resolution = [7, 12]
        self.psi0 = -0.3
        self.r = -0.3
        self.t = 0
        self.v = 1
        self.dt = 0.1

        # If there are additional arguments provided, set them as attributes
        # Possibly overwriting the default values
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.xRes = nx * self.micro_resolution[0]
        self.yRes = ny * self.micro_resolution[1]

        a0 = 2 * np.pi * 2 / np.sqrt(3)
        self.a = a0 * np.array([[1, 0], [1 / 2, np.sqrt(3) / 2], [1 / 2, -np.sqrt(3) / 2]])
        self.q = np.array([[np.sqrt(3) / 2, -1 / 2], [0, 1], [-np.sqrt(3) / 2, -1 / 2]])

        # Set the number of reciprocal modes
        self.number_of_reciprocal_lattice_modes = 3
        self.number_of_primary_reciprocal_lattice_modes = 3

        # Set the grid
        self.dx = a0 / self.micro_resolution[0]
        self.dy = np.sqrt(3) * a0 / self.micro_resolution[1]

        self.type_of_evolution = kwargs.get('type_of_evolution', 'conserved')
        if self.type_of_evolution == 'unconserved':
            self.psi0_proto, self.A_proto = self.calc_proto_amplitudes_unconserved()
        else:
            if self.type_of_evolution != 'conserved':
                tool_print_in_color('Warning: type_of_evolution should be either conserved or unconserved. Setting to conserved.', 'red')
            self.A_proto = self.calc_proto_amplitudes_conserved()   

        # Set the elastic constants
        self.el_lambda_proto = 3 * self.A_proto ** 2
        self.el_mu_proto = 3 * self.A_proto ** 2
        self.el_gamma_proto = 0

        bool_is_for_properties_calculation = kwargs.get('for_properties_calculation', False)

        if not bool_is_for_properties_calculation:
            tool_print_in_color('Initiating a 2D triangular PFC model.', 'green')
            kwargs.pop('for_properties_calculation', None)
            pfc = PhaseFieldCrystal2DTriangular(1,1,for_properties_calculation=True, type_of_evolution=self.type_of_evolution, **kwargs)
            final_strain, self.psi0, self.A, self.el_lambda, self.el_mu, self.el_gamma = pfc.calc_strained_amplitudes()     

            # Calculate S1111_ref
            dxxxpsi = pfc.ifft(pfc.dif[0] * pfc.dif[0] * pfc.dif[0] * pfc.fft(pfc.psi)).real
            dxyypsi = pfc.ifft(pfc.dif[0] * pfc.dif[1] * pfc.dif[1] * pfc.fft(pfc.psi)).real
            dyyypsi = pfc.ifft(pfc.dif[1] * pfc.dif[1] * pfc.dif[1] * pfc.fft(pfc.psi)).real

            self.S111111_ref = np.mean(dxxxpsi ** 2)
            self.Skkllmm_ref = np.mean(dxxxpsi ** 2 + 3 * dxxxpsi*dxyypsi + 3 * dxyypsi**2 + dyyypsi**2)

        else:
            self.A = self.A_proto
            self.el_lambda = self.el_lambda_proto
            self.el_mu = self.el_mu_proto
            self.el_gamma = self.el_gamma_proto
            
        self.eta0 = np.array([self.A, self.A, self.A])

        # Initialize the BaseSystem
        kwargs['xRes'] = self.xRes
        kwargs['yRes'] = self.yRes
        kwargs['dx'] = self.dx
        kwargs['dy'] = self.dy
        kwargs['dt'] = self.dt
        super().__init__(self.dim,  **kwargs)
        
        # Set the a0
        self.a0 = a0
        self.bool_has_defined_length_scale = True
        
        if not bool_is_for_properties_calculation:
            self.conf_apply_distortion([[final_strain,0],[0,final_strain]], update_q_and_a_vectors=True)
            self.a0 = self.a0 *(1+final_strain)



    def calc_proto_amplitudes_conserved(self):
        """Calculates the proto-amplitudes for the system.

        Args:
            None
        
        Returns:
            The proto-amplitudes for the system.
        """

        psi0 = self.psi0
        r = self.r
        t = self.t
        v = self.v
        
        #Testing which of the three amplitudes give the lowest free energy
        A = 0
        A_tmp = 1 / (15 * v)*(-t -3 * v * psi0 - np.sqrt(t**2 - 15 * v * r - 24 * t * v * psi0 - 36 * v ** 2 * psi0 ** 2)) 
        if self.calc_free_energy_from_proto_amplitudes(psi0, A_tmp) < self.calc_free_energy_from_proto_amplitudes(psi0, A):
            A = A_tmp
        A_tmp = 1 / (15 * v)*(-t -3 * v * psi0 + np.sqrt(t**2 - 15 * v * r - 24 * t * v * psi0 - 36 * v ** 2 * psi0 ** 2)) 
        if self.calc_free_energy_from_proto_amplitudes(psi0, A_tmp) < self.calc_free_energy_from_proto_amplitudes(psi0, A):
            A = A_tmp
        return A

    def calc_proto_amplitudes_unconserved(self):
        """Calculates the proto-amplitudes for the system.

        Args:
            None

        Returns:
            The proto-amplitudes for the system.
        """

        psi0 = 0
        A = 0
        free_energy = self.calc_free_energy_from_proto_amplitudes(psi0, A)

        for psi00 in np.linspace(-1,1,10):
            for A0 in np.linspace(0.01,1,5):
                [psi0_tmp, A_tmp] = fsolve(self.calc_proto_amplitude_equations_unconserved, [psi00, A0])
                free_energy_tmp = self.calc_free_energy_from_proto_amplitudes(psi0_tmp, A_tmp)

                if free_energy_tmp <= free_energy:
                    psi0 = psi0_tmp
                    A = A_tmp
                    free_energy = free_energy_tmp

        return psi0, A

    def calc_proto_amplitude_equations_unconserved(self,vars):
        """Calculates the equations for the proto-amplitudes for the system in the case of conserved dynamics

        Args:
            vars: The proto-amplitudes for the system.

        Returns:
            The equations for the proto-amplitudes for the system.
        """
        r = self.r
        t = self.t
        v = self.v
        
        psi0, A = vars
        eq1 = 12*A**3*self.v + 6*A**2*(self.t + 3*self.v*psi0) + psi0*(1 + self.r + self.t*psi0 + self.v*psi0**2)
        eq2 = self.r + 15*A**2*self.v + 2*A*(self.t + 3*self.v*psi0) + psi0*(2*self.t + 3*self.v*psi0) #*A not necessary  

        return [eq1, eq2]

    def calc_free_energy_from_proto_amplitudes(self, psi0, A):
        """Calculates the free energy of the system from the proto-amplitudes.

        Args:
            psi0: The average value of psi.
            A: The proto-amplitude.

        Returns:
            The free energy of the system.
        """
        r = self.r
        t = self.t
        v = self.v

        return np.pi**2/(3*np.sqrt(3))*(270*A**4*v + 48*A**3*( t + 3*v*psi0) + psi0**2*(6 + 6*r + 4*t*psi0 + 3*v*psi0**2) + 36*A**2*( r + psi0*(2*t + 3*v*psi0)))

    def calc_L_f(self):
        """Calculates the L operator in Fourier space.

        Args:
            None
        Returns:
            The L operator in Fourier space.
        """
        k2 = self.calc_k2()
        return 1 - k2

    def calc_L_sum_f(self):
        """Calculates the sum of the L operators in Fourier space. Needed for stress calculation functions.

        Args:
            None
        Returns:
            The L operator in Fourier space.
        """
        return 1
    


from typing import Optional
import numpy as np
import matplotlib
from comfit.tool.tool_complete_field import tool_complete_field


########################################################################################################################
################# FUNCTION NOT YET PORTED TO PLOTLY #####################################################################
########################################################################################################################

def plot_angle_field_in_plane_plotly(
        self,
        angle_field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        **kwargs
    ) -> matplotlib.axes.Axes:
    """Plots the angle field in a plane.

    Args:
        angle_field (numpy.ndarray): The angle field to be plotted.
        normal_vector (array-like, optional): The normal vector of the plane. Default is [0,1,0].
        position (array-like, optional): The position of the plane. Default is the middle of the system.
        **kwargs: Keyword arguments for the plot, see https://comfitlib.com/Plotting/.

    Returns:
        plotly.graph_objects.Figure: The figure containing the plot.
    """

    print('\033[93mWarning: The plot_angle_field_in_plane function is not yet supported with plotly.\033[0m')
    pass    

    # # Check if the vector field is complex
    # if np.iscomplexobj(angle_field):
    #     print("\033[91mWarning: the angle vector field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
    #     print('Max imaginary part: ', np.max(np.imag(angle_field)))
    #     angle_field = np.real(angle_field)

    # # Extend the field if not a complete array is given
    # angle_field = tool_complete_field(self, angle_field)

    # complex_field = np.exp(1j * angle_field)

    # return self.plot_complex_field_in_plane(complex_field, normal_vector=normal_vector, position=position, **kwargs)



import numpy as np
import plotly.graph_objects as go

from typing import Union

from comfit.tool import tool_complete_field
from comfit.tool import tool_set_plot_axis_properties_matplotlib
from comfit.tool import tool_set_plot_axis_properties_plotly
    
from comfit.tool import tool_colormap_bluewhitered, tool_colormap_sunburst, tool_complete_field

def plot_field_plotly(self, field: np.ndarray, **kwargs) -> go.Figure:
    """Plots the given (real) field.
    
    Args:
        field (array-like): The field to be plotted.
        **kwargs: Keyword arguments for the plot.
            See https://comfitlib.com/ClassBaseSystem/ 
            for a full list of keyword arguments.
    
    Returns:
        The figure containing the plot (plotly.graph_objects.Figure).
    """

    if field.dtype == bool:
        field = field.astype(float)

    field_is_nan = False
    if np.all(np.isnan(field)):
        field_is_nan = True

    if field.ndim == self.dim+1:
        print("\033[91mWarning - in plot_field: the provided field seems to be an order parameter containing several fields. Only the zeroth component will be plotted.\033[0m")
        field = field[0]

    # Check if the vector field is complex
    if np.iscomplexobj(field):
        print("\033[91mWarning - in plot_field: the provided field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
        print('Max imaginary part: ', np.max(np.imag(field)))
        field = np.real(field)

    fig = kwargs.get('fig', go.Figure())

    # Kewyord arguments
    colorbar = kwargs.get('colorbar', True)
    axis_equal = kwargs.get('axis_equal', None)

    # Extend the field if not a complete array is given
    field = tool_complete_field(self, field)

    kwargs['plot_is_3D'] = False

    # Colormap
    colormap = kwargs.get('colormap', 'Viridis')

    # Field limits
    field_min = np.min(field)
    field_max = np.max(field)

    if 'vlim' in kwargs:
        vlim = kwargs['vlim']
        vmin = vlim[0]
        vmax = vlim[1]
    else:
        vmin = field_min
        vmax = field_max
        if 'vlim_symmetric' in kwargs:
            if kwargs['vlim_symmetric']:
                vmax = max(abs(vmin), abs(vmax))
                vmin = -vmax

    ###############################################################
    ###################### DIMENSION: 1 ###########################
    ###############################################################

    if self.dim == 1:

        # Keyword arguments particular to the 1D case
        kwargs['grid'] = kwargs.get('grid', True)

        if axis_equal is None:
            kwargs['axis_equal'] = False

        trace = go.Scatter(
            x=self.x/self.a0,
            y=field,
            mode='lines',
            name='',
            hovertemplate='x: %{x:.1f} a₀<br>field: %{y:.1f}'
        )
        
        if not field_is_nan:
            fig.add_trace(trace)

    ###############################################################
    ###################### DIMENSION: 2 ###########################
    ###############################################################

    if self.dim == 2:
        
        # Keyword arguments particular to the 2D case
        kwargs['grid'] = kwargs.get('grid', False)

        X = kwargs.get('X', None)
        Y = kwargs.get('Y', None)

        if X is None or Y is None:
            X, Y = np.meshgrid(self.x, self.y, indexing='ij')
            
        trace = go.Heatmap(
            x=X.flatten()/self.a0,
            y=Y.flatten()/self.a0,
            z=field.flatten(),
            zmin=vmin,
            zmax=vmax,
            colorscale=colormap, 
            showscale=colorbar,
            zsmooth='best',
            hovertemplate='x: %{x:.2f} a₀<br>y: %{y:.2f} a₀<br> field: %{z:.2f}',
            name=''
        )

        if not field_is_nan:
            fig.add_trace(trace)

    ###############################################################
    ###################### DIMENSION: 3 ###########################
    ###############################################################
    elif self.dim == 3:

        kwargs['plot_is_3D'] = True

        # Keyword arguments particular to the 3D case

        number_of_layers = kwargs.get('number_of_layers', 1)
        alpha = kwargs.get('alpha', 0.5)
        

        if 'layer_values' in kwargs:
            layer_values = np.concatenate([[-np.inf], kwargs['layer_values'], [np.inf]])
        else: 
            layer_values = np.linspace(vmin, vmax, number_of_layers + 2)

        if 'colorbar' in kwargs:
            colorbar = kwargs['colorbar']
        else:
            colorbar = True

        #Plotting the layers
        X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')

        if not field_is_nan:
            for layer_value in layer_values[1:-1]:
                
                trace = go.Isosurface(
                    x=X.flatten()/self.a0,
                    y=Y.flatten()/self.a0,
                    z=Z.flatten()/self.a0,
                    value = field.flatten(),
                    isomin = layer_value,
                    isomax = layer_value,
                    cmin = vmin,
                    cmax = vmax,
                    surface=dict(count=3),  # Ensuring only one surface is shown
                    colorscale='Viridis',
                    opacity=alpha,
                    showscale=bool(layer_value == layer_values[1])
                )

                fig.add_trace(trace)
    
    kwargs['fig'] = fig
    tool_set_plot_axis_properties_plotly(self,**kwargs)
    return fig


import numpy as np
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.colors as mcolors


def tool_colormap_angle(pyplot=False):
    # Create hues ranging from cyan to red and back to cyan
    # Note: In the HSV color space, cyan is at 0.5 and red is at 0.
    hues = np.linspace(0.5, 0, 128).tolist() + np.linspace(1, 0.5, 128).tolist()

    # Reverse the order of the hues
    hues = hues[::-1]

    # Convert these hues to RGB colors with full saturation and value
    colors = [mcolors.hsv_to_rgb([h, 1, 1]) for h in hues]

    colormap = LinearSegmentedColormap.from_list("custom_hsv", colors)

    if pyplot:
         # # Convert the colormap to Plotly format
        custom_hsv = []
        for i in range(colormap.N):
            rgb = colormap(i)[:3]
            custom_hsv.append([i / (colormap.N - 1), f'rgb({int(rgb[0]*255)}, {int(rgb[1]*255)}, {int(rgb[2]*255)})'])
        return custom_hsv
    else:
        return colormap


def tool_colormap_bluewhitered():
    # Create hues ranging from cyan to red and back to cyan
    # Define the blue-white-red colormap
    colors = [(0, 'blue'), (0.5, 'white'), (1, 'red')]
    return mcolors.LinearSegmentedColormap.from_list('blue_white_red', colors)

def tool_colormap_sunburst():
    cm_data = [[0.24, 0.15, 0.66],
                [0.20, 0.48, 0.99],
                [0.07, 0.74, 0.72],
                [0.79, 0.75, 0.15],
                [0.9769,0.9839,0.0805]]
    return LinearSegmentedColormap.from_list('sunburst', cm_data)



import numpy as np
import plotly.graph_objects as go
import matplotlib.colors as mcolors
import matplotlib.cm as cm
import scipy as sp
from skimage.measure import marching_cubes
from comfit.tool import tool_complete_field
from comfit.tool import tool_colormap_angle
from comfit.tool import tool_set_plot_axis_properties_plotly


def plot_complex_field_plotly(self, complex_field: np.ndarray, **kwargs) -> go.Figure:
    """
    Plot a complex field.

    Args:
        complex_field (numpy.ndarray): The complex field to plot.
        ax (matplotlib.axes.Axes, optional): The matplotlib axes on which to plot the field.
            If not provided, a new 3D axes will be created.
    
    Returns:
        go.Figure: The figure containing the plot.
    """

    fig = kwargs.get('fig', go.Figure())

    # Kewyord arguments
    colorbar = kwargs.get('colorbar', True)

    # Extend the field if not a complete array is given
    complex_field = tool_complete_field(self, complex_field)

    # Calculate the magnitude and phase of the complex field
    rho = np.abs(complex_field)
    theta = np.angle(complex_field)

    kwargs['plot_is_3D'] = False

    ###############################################################
    ###################### DIMENSION: 1 ###########################
    ###############################################################

    if self.dim == 1:


        # Keyword arguments particular to the 1D case
        grid = kwargs.get('grid', False)
        
        axis_equal = kwargs.get('axis_equal',False)

        if fig == None:
            fig = go.Figure()

        # Color in the graph based on the argument of the complex field
        blend_factor=0.3 # The degree to which the color is blended with white
        cmap = tool_colormap_angle()

        color = (1-blend_factor)*np.array(cmap((theta[0] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
        color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')

        fig.add_trace(go.Scatter(x=[self.xmin/self.a0,(self.xmin+self.dx/2)/self.a0], 
                        y=[rho[0],(rho[0]+rho[1])/2],
                        mode='lines',
                        line=dict(color='rgba(0,0,0,0)'),
                        fill='tozeroy',
                        showlegend=False,
                        hoverinfo='skip',
                        fillcolor=color_str))

        for i in range(1,self.xRes-1):
            color = (1-blend_factor)*np.array(cmap((theta[i] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
            color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')
            fig.add_trace(go.Scatter(x=[(self.x[i]-self.dx/2)/self.a0,self.x[i]/self.a0], 
                        y=[(rho[i]+rho[i-1])/2,rho[i]],
                        mode='lines',
                        line=dict(color='rgba(0,0,0,0)'),
                        fill='tozeroy',
                        showlegend=False,
                        hoverinfo='skip',
                        fillcolor=color_str))

            color = (1-blend_factor)*np.array(cmap((theta[i] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
            color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')
            fig.add_trace(go.Scatter(x=[self.x[i]/self.a0,(self.x[i]+self.dx/2)/self.a0], 
                        y=[rho[i],(rho[i]+rho[i+1])/2],
                        mode='lines',
                        line=dict(color='rgba(0,0,0,0)'),
                        fill='tozeroy',
                        showlegend=False,
                        hoverinfo='skip',
                        fillcolor=color_str))

        color = (1-blend_factor)*np.array(cmap((theta[-1] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
        color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')
        fig.add_trace(go.Scatter(x=[(self.xmax-1.5*self.dx)/self.a0,(self.xmax-self.dx)/self.a0], 
                        y=[(rho[-1]+rho[-2])/2,rho[-1]],
                        mode='lines',
                        line=dict(color='rgba(0,0,0,0)'),
                        fill='tozeroy',
                        showlegend=False,
                        hoverinfo='skip',
                        fillcolor=color_str))


        fig.add_trace(go.Scatter(
            x=self.x/self.a0,
            y=rho,
            mode='lines',
            showlegend=False,
            customdata=np.stack((theta/np.pi, rho), axis=-1),
            hovertemplate='x: %{x:.2f} a₀<br>θ: %{customdata[0]:.2f} π<br>ρ: %{customdata[1]:.2e}',
            name='',
            line=dict(color='black')
        ))

        if colorbar:
            custom_colormap = tool_colormap_angle(pyplot=True)
            fig.add_trace(go.Scatter(
            x=[None], y=[None], mode='markers',
            showlegend=False,
            marker=dict(
                colorscale=custom_colormap,
                cmin=-np.pi, cmax=np.pi,
                colorbar=dict(
                tickvals=[-np.pi, -2*np.pi/3, -np.pi/3, 0, np.pi/3, 2*np.pi/3, np.pi],
                ticktext=['-π', '-2π/3', '-π/3', '0', 'π/3', '2π/3', 'π']
                )
            ),
            hoverinfo='none'
            ))
        
    ###############################################################
    ###################### DIMENSION: 2 ###########################
    ###############################################################

    elif self.dim == 2:
        # Keyword arguments particular to the 2D case
        plot_method = kwargs.get('plot_method', 'phase_angle')

        axis_equal = kwargs.get('axis_equal',True)

        # Create a meshgrid
        X, Y = np.meshgrid(self.x, self.y, indexing='ij')

        if plot_method == '3Dsurface':
        
            print("\033[91mWarning: 3D surface plot not yet implemented for Plotly.\033[0m")
            pass

        elif plot_method == 'phase_angle':
            

            # Keyword arguments particular to the phase angle plot
            grid = kwargs.get('grid', False)

            rho_normalized = rho / np.max(rho)

            custom_colormap = tool_colormap_angle()
            
            norm = mcolors.Normalize(vmin=-np.pi, vmax=np.pi)
            colormap = cm.ScalarMappable(norm=norm, cmap=custom_colormap)

            # Create an RGBA image array
            rgba_image = colormap.to_rgba(theta)

            # Apply the spatial opacity
            for i in range(rgba_image.shape[0]):
                for j in range(rgba_image.shape[1]):
                    for k in range(3):
                        rgba_image[i, j, k] = 1-(1-rgba_image[i, j, k])*rho_normalized[i, j]  # Set alpha channel

            rgba_image = np.transpose(rgba_image, (1, 0, 2))

            # Convert RGBA image to a format Plotly can understand
            image_data = (rgba_image * 255).astype(np.uint8)


            if fig == None:
                fig = go.Figure()

            fig.add_trace(go.Image(z=image_data, 
                            dx=self.dx/self.a0, 
                            dy=self.dy/self.a0, 
                            x0=self.xmin/self.a0, 
                            y0=self.ymin/self.a0,
                            hovertemplate='x: %{x:.2f} a₀<br>y: %{y:.2f} a₀<br>θ: %{customdata[0]:.2f} π<br>ρ: %{customdata[1]:.2e}',
                            customdata=np.stack((np.transpose(theta/np.pi), np.transpose(rho)), axis=-1),
                            name=''
                            ))  


                # fig.add_trace(go.Heatmap(
                #     x=X.flatten()/self.a0,
                #     y=Y.flatten()/self.a0,
                #     z=theta.flatten(),
                #     zmin=-np.pi,
                #     zmax=np.pi,
                #     colorscale=tool_colormap_angle(pyplot=True),
                #     colorbar=dict(
                #             tickvals=[-np.pi, -2*np.pi/3, -np.pi/3, 0, np.pi/3, 2*np.pi/3, np.pi],
                #             ticktext=['-π', '-2π/3', '-π/3', '0', 'π/3', '2π/3', 'π']
                #         )
                # ))

            if axis_equal:
                # Set axis to be equal
                fig.update_yaxes(
                    scaleanchor="x",
                    scaleratio=1,
                )

                fig.update_xaxes(
                    scaleanchor="y",
                    scaleratio=1,
                    ) 
    
        if colorbar:
            custom_colormap = tool_colormap_angle(pyplot=True)
            fig.add_trace(go.Scatter(
            x=[None], y=[None], mode='markers',
            showlegend=False,
            marker=dict(
                colorscale=custom_colormap,
                cmin=-np.pi, cmax=np.pi,
                colorbar=dict(
                tickvals=[-np.pi, -2*np.pi/3, -np.pi/3, 0, np.pi/3, 2*np.pi/3, np.pi],
                ticktext=['-π', '-2π/3', '-π/3', '0', 'π/3', '2π/3', 'π']
                )
            ),
            hoverinfo='none'
            ))
    ###############################################################
    ###################### DIMENSION: 3 ###########################
    ###############################################################

    elif self.dim == 3:
        kwargs['plot_is_3D'] = True

        axis_equal = kwargs.get('axis_equal',True)

        grid = kwargs.get('grid', True)

        plot_method = kwargs.get('plot_method', 'phase_blob')
        
        X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')

        rho_normalized = rho / np.max(rho)

        colormap = tool_colormap_angle()

        if plot_method == 'phase_angle':
        
            print('\033[91mWarning:\033[0m Plotly not yet implemented for 3D phase angle plots.')
            pass
        
        elif plot_method == 'phase_blob':
            # TODO: change the function so that it used plot_surface_matplotlib (Vidar 11.03.24)
            # Padding for the colorbar
            padding=0.2
            
            phase_blob_threshold = kwargs.get('phase_blob_threshold', 0.5)

            if np.nanmin(rho_normalized)<phase_blob_threshold<np.nanmax(rho_normalized):
                verts, faces, _, _ = marching_cubes(rho_normalized, phase_blob_threshold)

                # Calculate the centroids of each triangle
                centroids = np.mean(verts[faces], axis=1)

                # Assuming theta is defined on the same grid as rho
                x, y, z = np.mgrid[0:rho_normalized.shape[0], 0:rho_normalized.shape[1], 0:rho_normalized.shape[2]]

                # Flatten the grid for interpolation
                points = np.c_[x.ravel(), y.ravel(), z.ravel()]
                theta_values = theta.ravel()

                # Interpolate theta at the vertices positions
                reals = np.real(complex_field)
                imags = np.imag(complex_field)

                interpolation_method = kwargs.get('interpolation_method', 'linear')
                print("Interpolating points with method ", interpolation_method, ".")
                print("If this process is slow, consider passing 'interpolation_method='nearest' with the plot_complex_field function.")
                print("That will speed up the process, but the plot may look less smooth.")
                reals_faces = sp.interpolate.griddata(points, reals.ravel(), centroids, method='nearest')
                imags_faces = sp.interpolate.griddata(points, imags.ravel(), centroids, method='nearest')
                print("Interpolation done.")
                
                theta_faces = np.arctan2(imags_faces, reals_faces)

                # Normalize theta values for color mapping
                theta_faces_normalized = (theta_faces + np.pi) / (2*np.pi)

                # Map normalized theta values to colors
                colors = colormap(theta_faces_normalized)

                # Convert colors to 'rgba()' format required by Plotly
                colors = ['rgba({},{},{},{})'.format(*c) for c in (255*colors).astype(int)]

                # Create the mesh object
                x_new = (verts[:, 0] * self.dx + self.xmin) / self.a0
                y_new = (verts[:, 1] * self.dy + self.ymin) / self.a0
                z_new = (verts[:, 2] * self.dz + self.zmin) / self.a0

                mesh = go.Mesh3d(
                    x=x_new, 
                    y=y_new, 
                    z=z_new,
                    i=faces[:, 0], 
                    j=faces[:, 1], 
                    k=faces[:, 2],
                    facecolor=colors,  # Set color for each face
                    showscale=True
                )

                fig = go.Figure(data=[mesh])

    if colorbar:
        # Add a colorbar to the plot
        if self.dim == 1:
            pass
        elif self.dim == 2:
            pass
        elif self.dim == 3:
            # Generate the custom colormap
            custom_colormap = tool_colormap_angle(pyplot=True)

            fig.add_trace(go.Isosurface(
                x=[self.xmin/self.a0,self.xmax/self.a0], 
                y=[self.ymin/self.a0,self.ymax/self.a0], 
                z=[self.zmin/self.a0, self.zmax/self.a0],
                value=[0,0],
                cmin = -np.pi,
                cmax = np.pi,
                opacity=0,
                colorscale=custom_colormap,
                colorbar=dict(
                    tickvals=[-np.pi, -2*np.pi/3, -np.pi/3, 0, np.pi/3, 2*np.pi/3, np.pi],
                    ticktext=['-π', '-2π/3', '-π/3', '0', 'π/3', '2π/3', 'π']
                ),
                showscale=True
            ))


    kwargs['fig'] = fig
    kwargs['axis_equal'] = axis_equal
    tool_set_plot_axis_properties_plotly(self, **kwargs)
    return fig


import numpy as np
import plotly.graph_objects as go

def tool_set_plot_axis_properties_plotly(self, **kwargs):
    """Sets the properties of the axis for a plot.

    Args:
    kwargs: keyword arguments for the axis properties

    Returns:
    The axis object with the properties set.
    """
    # Create dictionaries to store the layout updates
    layout_updates = {}
    
    # For 3D plots, create dictionaries to store the scene updates
    scene_updates = {}
    xaxis_updates = {}
    yaxis_updates = {}
    zaxis_updates = {}

    ##### FIGURE #####
    fig = kwargs.get('fig', go.Figure())

    ##### SIZE #####
    size = kwargs.get('size', None)

    if size is not None:
        layout_updates['width'] = size[0]
        layout_updates['height'] = size[1]

    ##### PLOT NATURE #####
    row = kwargs.get('row', None)
    col = kwargs.get('col', None)

    plot_is_3D = kwargs.get('plot_is_3D', False)

    ##### GRID #####
    grid = kwargs.get('grid', True)
    layout_updates['xaxis_showgrid'] = grid
    layout_updates['yaxis_showgrid'] = grid

    ##### AXIS EQUAL #####
    axis_equal = kwargs.get('axis_equal', False if self.dim==1 else True)

    if axis_equal:
        if plot_is_3D:
            scene_updates['aspectmode'] = 'cube'
        else:
            layout_updates['yaxis'] = dict(scaleanchor='x', scaleratio=1)
    

    ##### TITLE #####
    title = kwargs.get('title', None)

    if title is not None:
        layout_updates['title'] = title

    ##### AXIS LIMITS #####
    # xlim is specified as a list
    xlim = [self.xmin/self.a0, (self.xmax-self.dx)/self.a0]
    if 'xmin' in kwargs:
        xlim[0] = kwargs['xmin'] / self.a0
    if 'xmax' in kwargs:
        xlim[1] = kwargs['xmax'] / self.a0
    if 'xlim' in kwargs:
        xlim = np.array(kwargs['xlim']) / self.a0

    # ylim is specified as a list if dim>1 else as None
    ylim = [self.ymin/self.a0, (self.ymax-self.dy)/self.a0] if self.dim > 1 else None
    if 'ymin' in kwargs:
        ylim[0] = kwargs['ymin'] / self.a0 if self.dim > 1 else kwargs['ymin']
    if 'ymax' in kwargs:
        ylim[1] = kwargs['ymax'] / self.a0 if self.dim > 1 else kwargs['ymax']
    if 'ylim' in kwargs:
        ylim = np.array(kwargs['ylim'])/self.a0 if self.dim > 1 else kwargs['ylim']

    # zlim is specified as a list if dim>2 else as None
    zlim = [self.zmin/self.a0, (self.zmax-self.dz)/self.a0] if self.dim > 2 else None
    if 'zmin' in kwargs:
        zlim[0] = kwargs['zmin'] / self.a0 if self.dim > 2 else kwargs['zmin']
    if 'zmax' in kwargs:
        zlim[1] = kwargs['zmax'] / self.a0 if self.dim > 2 else kwargs['zmax']
    if 'zlim' in kwargs:
        zlim = np.array(kwargs['zlim'])/self.a0 if self.dim > 2 else kwargs['zlim']

    if plot_is_3D:
        xaxis_updates['range'] = xlim
        yaxis_updates['range'] = ylim
        zaxis_updates['range'] = zlim

    else:
        layout_updates['xaxis_range'] = xlim
        layout_updates['yaxis_range'] = ylim

    ##### AXIS LABELS #####
    xlabel = kwargs.get('xlabel', 'x/a₀')
    ylabel = kwargs.get('ylabel', 'y/a₀' if self.dim > 1 else None)
    zlabel = kwargs.get('zlabel', 'z/a₀' if self.dim > 2 else None)

    if plot_is_3D:
        xaxis_updates['title'] = xlabel
        yaxis_updates['title'] = ylabel
        zaxis_updates['title'] = zlabel

    else: #plot is not 3D
        layout_updates['xaxis_title'] = xlabel

        if ylabel is not None:
            layout_updates['yaxis_title'] = ylabel


    ##### TICKS #####
    xticks = kwargs.get('xticks', None)  
    xticklabels = kwargs.get('xticklabels', None)

    yticks = kwargs.get('yticks', None)
    yticklabels = kwargs.get('yticklabels', None)

    zticks = kwargs.get('zticks', None)
    zticklabels = kwargs.get('zticklabels', None)

    
    if xticks is not None:
        xaxis_updates['tickvals'] = xticks
    if xticklabels is not None:
        xaxis_updates['ticktext'] = xticklabels

    if yticks is not None:
        yaxis_updates['tickvals'] = yticks
    if yticklabels is not None:
        yaxis_updates['ticktext'] = yticklabels

    
    if zticks is not None:
        zaxis_updates['tickvals'] = zticks
    if zticklabels is not None:
        zaxis_updates['ticktext'] = zticklabels

    ##### UPDATE SCENE #####
    if plot_is_3D:
        if xaxis_updates:
            scene_updates['xaxis'] = xaxis_updates

        if yaxis_updates:
            scene_updates['yaxis'] = yaxis_updates

        if zaxis_updates:
            scene_updates['zaxis'] = zaxis_updates

    else:
        if xaxis_updates:
            layout_updates['xaxis'] = xaxis_updates

        if yaxis_updates:
            layout_updates['yaxis'] = yaxis_updates


    ##### UPDATE LAYOUT #####
    fig.update_layout(layout_updates)
    if plot_is_3D:
        fig.update_layout(scene = scene_updates)

    
    


import numpy as np
import matplotlib.pyplot as plt
from comfit.tool.tool_complete_field import tool_complete_field
from comfit.tool.tool_set_plot_axis_properties_matplotlib import tool_set_plot_axis_properties_matplotlib

def plot_field_velocity_and_director_matplotlib(self, field, velocity, director, **kwargs):
    """Plot the fields, velocity, and director field in 2 dimensions

    Args:
        field (ndarray): The field to be plotted.
        velocity (ndarray): The velocity to be plotted.
        director (ndarray): The director to be plotted.
        **kwargs: Keyword arguments for the plot.
            See https://comfitlib.com/ClassBaseSystem/
            for a full list of keyword arguments.

    Returns:
        A tuple consisting of
            - The figure
            - The axes with the plotted field, velocity, and director. ax (Axes)

    Raises:
        Exception: If the dimension is other than 2.
    """
    if field.dtype == bool:
        field = field.astype(float)

    # Check if the vector field is complex
    if np.iscomplexobj(field):
        print(
                "\033[91mWarning: the provided field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
        print('Max imaginary part: ', np.max(np.imag(field)))
    field = np.real(field)

    # Check if an axis object is provided
    fig = kwargs.get('fig', plt.gcf())
    ax = kwargs.get('ax', None)

    # Kewyord arguments
    colorbar = kwargs.get('colorbar', True)

    # Extend the field if not a complete array is given
    field = tool_complete_field(self, field)

    if self.dim == 2:

        # Keyword arguments particular to the 2D case
        kwargs['grid'] = kwargs.get('grid', False)

        if ax == None:
            fig.clf()
            ax = plt.gcf().add_subplot(111)

        colormap = kwargs.get('colormap', 'viridis')

        if colormap == 'bluewhitered':
            colormap = tool_colormap_bluewhitered()

        elif colormap == 'sunburst':
            colormap = tool_colormap_sunburst()
        else:
            colormap = plt.get_cmap(colormap)

        X, Y = np.meshgrid(self.x, self.y, indexing='ij')
        pcm = ax.pcolormesh(X / self.a0, Y / self.a0, field, shading='gouraud', cmap=colormap)

        xlim = [self.xmin, self.xmax - self.dx]
        ylim = [self.ymin, self.ymax - self.dy]

        limits_provided = False
        if 'xlim' in kwargs:
            xlim = kwargs['xlim']
            limits_provided = True
        else:
            if 'xmin' in kwargs:
                xlim[0] = kwargs['xmin']
                limits_provided = True

            if 'xmax' in kwargs:
                xlim[1] = kwargs['xmax']
                limits_provided = True

        if 'ylim' in kwargs:
            ylim = kwargs['ylim']
            limits_provided = True
        else:
            if 'ymin' in kwargs:
                ylim[0] = kwargs['ymin']
                limits_provided = True

            if 'ymax' in kwargs:
                ylim[1] = kwargs['ymax']
                limits_provided = True
        if limits_provided:
            region_to_plot = np.zeros(self.dims).astype(bool)
            region_to_plot[(xlim[0] <= X) * (X <= xlim[1]) * (ylim[0] <= Y) * (Y <= ylim[1])] = True
            vlim = [np.min(field[region_to_plot]), np.max(field[region_to_plot])]

        else:
            vlim = [np.min(field), np.max(field)]

        # Set the value limitses
        if 'vlim' in kwargs:
            vlim = kwargs['vlim']
        else:
            if 'vmin' in kwargs:
                vlim[0] = kwargs['vmin']
            if 'vmax' in kwargs:
                vlim[1] = kwargs['vmax']

        if vlim[1] - vlim[0] < 1e-10:
            vlim = [vlim[0] - 0.05, vlim[1] + 0.05]

        pcm.set_clim(vmin=vlim[0], vmax=vlim[1])

        if 'vlim_symmetric' in kwargs:
            vlim_symmetric = kwargs['vlim_symmetric']
            if vlim_symmetric:
                cmax = abs(field).max()
                cmin = -cmax
                pcm.set_clim(vmin=cmin, vmax=cmax)

        colorbar = kwargs.get('colorbar', True)

        if colorbar:
            cbar = plt.colorbar(pcm, ax=ax)



        ax.streamplot(X.T, Y.T, (velocity[0]).T, (velocity[1]).T, color='w')
        ax.quiver(X, Y, director[0], director[1], headwidth=0, scale=50)
        ax.quiver(X, Y, -director[0], -director[1], headwidth=0, scale=50)
        ax.set_aspect('equal')

    else:
        raise Exception("This plotting function is currently only implemented in 2D! ")

    kwargs['fig'] = fig
    tool_set_plot_axis_properties_matplotlib(self, **kwargs)
    return fig, ax


from typing import Optional

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import scipy as sp
from skimage.measure import marching_cubes

from comfit.tool.tool_complete_field import tool_complete_field
from comfit.tool.tool_set_plot_axis_properties_matplotlib import tool_set_plot_axis_properties_matplotlib
from comfit.tool.tool_colormaps import tool_colormap_angle, tool_colormap_bluewhitered


def plot_field_in_plane_matplotlib(
        self,
        field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        **kwargs
    ) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    
    """Plots the field in a plane perpendicular to the given normal vector
    
    Uses scipy.interpolate.griddata and plt.plot_trisurf.

    Args:
        field (array-like): The field to be plotted.
        normal_vector (array-like, optional): The normal vector of the plane. Default is [0,1,0].
        position (array-like, optional): The position of the plane. Default is the middle of the system.
        **kwargs: Keyword arguments for the plot, see https://comfitlib.com/Plotting/. 
    
    Returns:
        tuple: A tuple containing:
            - matplotlib.figure.Figure: The figure containing the plot.
            - matplotlib.axes.Axes: The axes containing the plot
    """

    # print('\033[93mWarning: The plot_in_plane function is not yet supported with plotly.\033[0m')

    if self.dim != 3:
        raise Exception("The plot in plane function is only defined for 3D fields.")

    # Extend the field if not a complete array is given
    field = tool_complete_field(self, field)

    # Check if the vector field is complex
    if np.iscomplexobj(field):
        print("\033[91mWarning: the provided field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
        print('Max imaginary part: ', np.max(np.imag(field)))
        field = np.real(field)

    # Check if an axis object is provided
    fig = kwargs.get('fig', plt.gcf())
    ax = kwargs.get('ax', None)

    # Kewyord arguments
    colorbar = kwargs.get('colorbar', True)

    # Default values of position and normal vector
    if position is None:
        position = self.rmid

    if normal_vector is None:
        normal_vector=[0,1,0]

    if ax is None:
        ax = fig.add_subplot(111, projection='3d')

    colormap = kwargs.get('colormap', 'viridis')

    if colormap == 'angle':
        colormap = tool_colormap_angle()
    elif colormap == 'bluewhitered':
        colormap = tool_colormap_bluewhitered()
    else:
        colormap = plt.get_cmap(colormap)

    normal_vector = np.array(normal_vector)/np.linalg.norm(normal_vector)
    height_above_plane = (self.x-position[0])*normal_vector[0] + (self.y-position[1])*normal_vector[1] + (self.z-position[2])*normal_vector[2]

    verts, faces, _, _ = marching_cubes(height_above_plane, 0)

    # Calculate the centroids of each triangle
    centroids = np.mean(verts[faces], axis=1)

    # Assuming field is defined on the same grid as height_above_plane
    x, y, z = np.mgrid[0:height_above_plane.shape[0], 0:height_above_plane.shape[1], 0:height_above_plane.shape[2]]

    # Flatten the grid for interpolation
    points = np.c_[x.ravel(), y.ravel(), z.ravel()]
    field_values = field.ravel()

    # Interpolate field at the vertices positions
    field_verts = sp.interpolate.griddata(points, field_values, centroids, method='nearest')

    # Normalize field values for color mapping
    field_normalized = (field_verts - np.min(field_verts)) / (np.max(field_verts) - np.min(field_verts))

    # Map normalized field values to colors
    colors = colormap(field_normalized)

    ax.plot_trisurf((self.xmin+verts[:, 0]*self.dx)/self.a0,
                    (self.ymin+verts[:, 1]*self.dy)/self.a0,
                    faces,
                    (self.zmin+verts[:, 2]*self.dz)/self.a0,
                    facecolor=colors, antialiased=False)

    if colorbar:
        sm = plt.cm.ScalarMappable(cmap=colormap)  
        sm.set_clim(np.min(field_verts),np.max(field_verts))  
        cbar = plt.colorbar(sm, ax=ax, pad=0.2)

    kwargs['grid'] = kwargs.get('grid', True)
    kwargs['ax'] = ax
    tool_set_plot_axis_properties_matplotlib(self,**kwargs)

    return fig, ax




def plot_save_plotly(counter, fig, **kwargs):
    """
    Saves the current Plotly figure as a PNG file.

    Args:
        counter (int): A unique identifier for the plot image file.
        fig (plotly.graph_objects.Figure): The Plotly figure to save as an image.
        ID (str, optional): A unique identifier for the plot image file. Defaults to None.
    
    Returns: 
        None, saves the plot as a PNG file.
    """

    # Keyword arguments
    ID = kwargs.get('ID', None)

    if ID is None:
        filename = f'plot_{counter}.png'
    else:
        filename = f'plot_{counter}_{ID}.png'

    fig.write_image(filename)


import numpy as np
import plotly.graph_objects as go
import plotly.figure_factory as ff
from comfit.tool import tool_set_plot_axis_properties_plotly

def plot_nodes_plotly(self, nodes, **kwargs):
        """Plots the  nodes in the system.

        Args:
            nodes (list): A list of dictionaries representing nodes. The dicitonary must contain:
                                 - 'position': The position of the  node in the defect density array.
                                 The following keys are optional:
                                 - 'velocity': The velocity of the node.
                                 - 'charge': The charge of the node.
                                 - 'tangent_vector': The tangent of the node (3D only).
                                 - 'Burgers_vector': The Burgers vector of the node (for dislocations).
            -**kwargs: Keyword arguments for the plot.
                See https://comfitlib.com/ClassBaseSystem/
                for a full list of keyword arguments.
        
        Returns:
            The axes on which the  nodes are plotted. (matplotlib.axes.Axes: )
        """

        # Check if an axis object is provided
        fig = kwargs.get('fig', go.Figure())

        # Check if there are nodes to be plotted
        if not nodes:
            return fig

        if self.dim == 2:

            # Positions
            x_coords = []
            y_coords = []

            x_coords_pos = []
            y_coords_pos = []

            x_coords_neg = []
            y_coords_neg = []

            charge_given = 'charge' in nodes[0].keys()

            if charge_given:

                for node in nodes:
                    x_coords.append(node['position'][0])
                    y_coords.append(node['position'][1])

                    if node['charge'] > 0:
                        x_coords_pos.append(node['position'][0])
                        y_coords_pos.append(node['position'][1])

                    else:
                        x_coords_neg.append(node['position'][0])
                        y_coords_neg.append(node['position'][1])

            else: # No charge given
                for node in nodes:
                    x_coords_pos.append(node['position'][0])
                    y_coords_pos.append(node['position'][1])
                
                x_coords = x_coords_pos
                y_coords = y_coords_pos

            if charge_given:
                if len(x_coords_neg) > 0:
                    fig.add_trace(go.Scatter(x=np.array(x_coords_neg)/self.a0, 
                                            y=np.array(y_coords_neg)/self.a0, 
                                            mode='markers', 
                                            marker=dict(symbol='star', color='blue'),
                                            showlegend=False))

                if len(x_coords_pos) > 0:
                    fig.add_trace(go.Scatter(x=np.array(x_coords_pos)/self.a0, 
                                            y=np.array(y_coords_pos)/self.a0, 
                                            mode='markers', 
                                            marker=dict(symbol='cross', color='red'),
                                            showlegend=False))
            
            else: # If no charge is provided, plot all nodes as positive
                print("Hello", x_coords, y_coords)
                fig.add_trace(go.Scatter(x=np.array(x_coords)/self.a0, 
                                        y=np.array(y_coords)/self.a0, 
                                        mode='markers', 
                                        marker=dict(symbol='circle', color='black'),
                                        showlegend=False))
            
            # Velocities
            if 'velocity' in nodes[0].keys():
                velocity_scale = 0.1*min(self.size_x, self.size_y)/self.a0

                vx = []
                vy = []
                
                for node in nodes:
                    vx.append(node['velocity'][0])
                    vy.append(node['velocity'][1])
                
                vx = np.array(vx)
                vy = np.array(vy)

                vnorm = np.sqrt(np.max(vx**2 + vy**2))
                    
                vx = velocity_scale*vx/vnorm
                vy = velocity_scale*vy/vnorm

                fig1 = ff.create_quiver(x=np.array(x_coords)/self.a0, 
                                        y=np.array(y_coords)/self.a0, 
                                        u=vx, 
                                        v=vy, 
                                        line=dict(width=1, color='black'),
                                        scale=1,
                                        showlegend=False)
                    
                fig.add_traces(data=fig1.data)

            # Burgers vectors
            if 'Burgers_vector' in nodes[0].keys():
                Burgers_vector_scale = 0.1*min(self.size_x, self.size_y)/self.a0

                bvx = []
                bvy = []
                
                for node in nodes:
                    bvx.append(node['Burgers_vector'][0])
                    bvy.append(node['Burgers_vector'][1])
                
                bvx = np.array(bvx)
                bvy = np.array(bvy)

                bnorm = np.sqrt(np.max(bvx**2 + bvy**2))
                    
                bvx = Burgers_vector_scale*bvx/bnorm
                bvy = Burgers_vector_scale*bvy/bnorm

                fig1 = ff.create_quiver(x=np.array(x_coords)/self.a0, 
                                        y=np.array(y_coords)/self.a0, 
                                        u=bvx, 
                                        v=bvy, 
                                        line=dict(width=1, color='red'),
                                        scale=1,
                                        showlegend=False)
                    
                fig.add_traces(data=fig1.data)

        elif self.dim == 3:
            # Plotting options
            quiver_scale = 2 # The scale of the quiver arrows

            x_coords = []
            y_coords = []
            z_coords = []

            tx = []
            ty = []
            tz = []

            vx = []
            vy = []
            vz = []


            for node in nodes:
                x_coords.append(node['position'][0])
                y_coords.append(node['position'][1])
                z_coords.append(node['position'][2])

                tx.append(node['tangent_vector'][0])
                ty.append(node['tangent_vector'][1])
                tz.append(node['tangent_vector'][2])

                vx.append(node['velocity'][0])
                vy.append(node['velocity'][1])
                vz.append(node['velocity'][2])

            tx = np.array(tx)
            ty = np.array(ty)
            tz = np.array(tz)

            vx = np.array(vx)
            vy = np.array(vy)
            vz = np.array(vz)

            if not len(vx) == 0:
                v2 =vx**2 + vy**2 + vz**2
                v_norm = np.sqrt(max(v2))
            else:
                v_norm = 1

            if len(x_coords) > 0:
                #ax.scatter(x_coords, y_coords, z_coords, marker='o', color='black')\
                fig.add_trace(go.Scatter(x=x_coords/self.a0, y=y_coords/self.a0, mode='markers', marker=dict(symbol='circle', color='black')))

                fig.add_trace(go.Cone(x=x_coords/self.a0, y=y_coords/self.a0, z=z_coords/self.a0, u=tx, v=ty, w=tz, colorscale='Blues', sizemode='scaled', sizeref=0))

                fig.add_trace(go.Cone(x=x_coords/self.a0, y=y_coords/self.a0, z=z_coords/self.a0, u=vx, v=vy, w=vz, colorscale='Greens', sizemode='scaled', sizeref=0))
            

        kwargs['fig'] = fig
        tool_set_plot_axis_properties_plotly(self, **kwargs)

        return fig


"""
This module provides tool for saving Matplotlib plots as img and creating animations from these img using MoviePy.


Functions:

- tool_save_plot_matplotlib: Saves a Matplotlib plot as an image file.
- tool_save_plot_plotly: Saves a Plotly figure as an image file.
- tool_make_animation: Creates an animation from a sequence of image files and saves it as a video file.
"""


import matplotlib.pyplot as plt

try:
    from moviepy import ImageSequenceClip
except ImportError:
    from moviepy.editor import ImageSequenceClip
    
import os
from datetime import datetime
from PIL import Image


def tool_make_animation_movie(counter, name=None, fps=24):
    """

    Creates an animation from a series of plot img and saves it as an MP4 video file.


    Args:
        counter (int): The number of plot img to include in the animation.
        name (str, optional): The filename for the output video. Defaults to today's date followed by ' - output_video.mp4'.
        fps (int, optional): The frames per second for the video. Defaults to 24.
    
    Returns:
        None, saves the video file.
    """
    
    if name is None:
        name = datetime.now().strftime("%y%m%d_%H%M") + ' - output_video.mp4'
    else:
        name = datetime.now().strftime("%y%m%d_%H%M") + ' - ' + name + '.mp4'

    # List of saved plot filenames
    image_files = [f'plot_{counter}.png' for counter in range(counter+1)]

    # Create the video clip from the image files
    video_clip = ImageSequenceClip(image_files, fps=fps)

    # Save the video
    video_clip.write_videofile(name)

    # Delete the png files
    for file in image_files:
        os.remove(file)


def tool_make_animation_gif(counter, 
                            name=None, 
                            fps=24,
                            ID=None):
    """
    Creates an animation from a series of plot img and saves it as a GIF file.

    Args:
        counter (int): The number of plot img to include in the animation.
        name (str, optional): The filename for the output video. Defaults to today's date followed by ' - output_video.mp4'.
        fps (int, optional): The frames per second for the video. Defaults to 24.
        ID (str, optional): A unique identifier for the plot image files. Defaults to None.

    Returns:
        None, saves the GIF file.
    """

    if name is None:
        name = datetime.now().strftime("%y%m%d_%H%M%S") + ' - output_animation.gif'
    else:
        name = datetime.now().strftime("%y%m%d_%H%M%S") + ' - ' + name + '.gif'

    if ID is None:
        image_files = [f'plot_{counter}.png' for counter in range(counter+1)]
    else:
        image_files = [f'plot_{counter}_{ID}.png' for counter in range(counter+1)]
    images = [Image.open(image_file) for image_file in image_files]

    # Find the smallest image dimensions
    min_width = min(image.width for image in images)
    min_height = min(image.height for image in images)

    # Crop images if necessary and issue a warning
    cropped_images = []
    for image in images:
        if image.width != min_width or image.height != min_height:
            print(f"Warning: Image sizes inconsistent. Cropping image {image.filename} to {min_width}x{min_height}. Consider reviewing how plots are saved mid loop. ")
            image = image.crop((0, 0, min_width, min_height))
        cropped_images.append(image)

    # Save the images as a GIF
    cropped_images[0].save(name, save_all=True, append_images=cropped_images[1:], duration=1000/fps, loop=0)

    # Delete the png files
    for file in image_files:
        os.remove(file)




import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from comfit.plot.plot_complex_field_matplotlib import plot_complex_field_matplotlib
from comfit.tool import tool_complete_field

####################################################################################################
########### NOT YET PORTED TO PLOTLY ###############################################################
####################################################################################################

def plot_angle_field_plotly(self, angle_field: np.ndarray, **kwargs) -> matplotlib.axes.Axes:
    """Plot the angle field.

    Args:
        field (array-like): The angle field values.
        ax (matplotlib.axes.Axes, optional): The axes to plot the angle field on. If not provided, a new subplot will be created.
    
    Returns:
        The axes containing the plot. (matplotlib.axes.Axes)
    """

    print("\033[91mWarning: The plot_angle_field_plotly function is not yet implemented.\033[0m")
    pass

    # # Check if the vector field is complex
    # if np.iscomplexobj(angle_field):
    #     print("\033[91mWarning: the angle vector field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
    #     print('Max imaginary part: ', np.max(np.imag(angle_field)))
    #     angle_field = np.real(angle_field)

    # # Extend the field if not a complete array is given
    # angle_field = tool_complete_field(self, angle_field)

    # # Normalize around 0
    # angle_field = np.mod(angle_field + np.pi, 2 * np.pi) - np.pi        

    # if self.dim == 1:
    #     if 'vlim' in kwargs:
    #         vlim = kwargs['vlim']
    #     else:
    #         kwargs['vlim'] = [-np.pi, np.pi]
    #         kwargs['yticks'] = [-np.pi, -2 * np.pi / 3, -np.pi / 3, 0, np.pi / 3, 2 * np.pi / 3, np.pi]
    #         kwargs['yticklabels'] = [r'$-\pi$', r'$-2\pi/3$', r'$-\pi/3$', r'$0$', r'$\pi/3$', r'$2\pi/3$', r'$\pi$']

        
    #     return plot_field_matplotlib(self, angle_field, **kwargs)
    
    # elif self.dim > 1:
    #     complex_field = np.exp(1j * angle_field)

    #     kwargs['plot_method'] = 'phase_angle'

    #     return plot_complex_field_matplotlib(self, complex_field, **kwargs)




def tool_add_spacing_3D(X,Y,Z,U,V,W,spacing):
        X = X[::spacing, ::spacing, ::spacing]
        Y = Y[::spacing, ::spacing, ::spacing]
        Z = Z[::spacing, ::spacing, ::spacing]
        U = U[::spacing, ::spacing, ::spacing]
        V = V[::spacing, ::spacing, ::spacing]
        W = W[::spacing, ::spacing, ::spacing]
        return X,Y,Z,U,V,W  


import numpy as np
import matplotlib.pyplot as plt

def plot_nodes_matplotlib(self, nodes, **kwargs):
    """
    Plots the nodes.

    Args:
        nodes: The nodes to plot.
        **kwargs: Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/ for a full list of keyword arguments.  
            
    Returns: 
        The axes containing the plot. (matplotlib.axes.Axes)
    """    

    # Check if an axis object is provided
    fig = kwargs.get('fig', plt.gcf())
    ax = kwargs.get('ax', None)

    if self.dim == 2:

        if ax == None:
            fig.clf()
            ax = fig.add_subplot(111)

        x_coords = []
        y_coords = []

        # vx_coords = []
        # vy_coords = []

        Bx_coords = []
        By_coords = []

        for node in nodes:

            x_coords.append(node['position'][0])
            y_coords.append(node['position'][1])
            # vx_coords.append(vortex['velocity'][0])
            # vy_coords.append(vortex['velocity'][1])
            Bx_coords.append(node['Burgers_vector'][0])
            By_coords.append(node['Burgers_vector'][1])

        x_coords = np.array(x_coords)
        y_coords = np.array(y_coords)

        # print(x_coords_pos,y_coords_pos)
        # print(x_coords_neg,y_coords_neg)
        ax.scatter(x_coords/self.a0, y_coords/self.a0, marker='o', color='black')

        # ax.quiver(x_coords, y_coords, vx_coords, vy_coords, color='black')
        ax.quiver(x_coords/self.a0, y_coords/self.a0, Bx_coords, By_coords, color='red')
        
        kwargs['ax'] = ax
        self.tool_set_plot_axis_properties_matplotlib(**kwargs)

    elif self.dim == 3:
        # Plotting options
        quiver_scale = 2 # The scale of the quiver arrows

        if ax == None:
            plt.clf()
            ax = plt.gcf().add_subplot(111, projection='3d')

        x_coords = []
        y_coords = []
        z_coords = []

        tx = []
        ty = []
        tz = []

        # vx = []
        # vy = []
        # vz = []

        Bx = []
        By = []
        Bz = []


        for node in nodes:
            x_coords.append(node['position'][0])
            y_coords.append(node['position'][1])
            z_coords.append(node['position'][2])

            tx.append(node['tangent_vector'][0])
            ty.append(node['tangent_vector'][1])
            tz.append(node['tangent_vector'][2])

            # vx.append(node['velocity'][0])
            # vy.append(node['velocity'][1])
            # vz.append(node['velocity'][2])

            Bx.append(node['Burgers_vector'][0])
            By.append(node['Burgers_vector'][1])
            Bz.append(node['Burgers_vector'][2])

        tx = np.array(tx)
        ty = np.array(ty)
        tz = np.array(tz)

        # vx = np.array(vx)
        # vy = np.array(vy)
        # vz = np.array(vz)

        Bx = np.array(Bx)
        By = np.array(By)
        Bz = np.array(Bz)

        # if not len(vx) == 0:
        #     v2 =vx**2 + vy**2 + vz**2
        #     v_norm = np.sqrt(max(v2))
        # else:
        #     v_norm = 1

        if not len(Bx) == 0:
            B2 =Bx**2 + By**2 + Bz**2
            B_norm = np.sqrt(max(B2))
        else:
            B_norm = 1

        #ax.scatter(x_coords, y_coords, z_coords, marker='o', color='black')
        ax.quiver(x_coords, y_coords, z_coords, quiver_scale*tx, quiver_scale*ty, quiver_scale*tz, color='blue')
        # ax.quiver(x_coords, y_coords, z_coords, quiver_scale*vx/v_norm, quiver_scale*vy/v_norm, quiver_scale*vz/v_norm, color='green')
        ax.quiver(x_coords, y_coords, z_coords, quiver_scale*Bx/B_norm, quiver_scale*By/B_norm, quiver_scale*Bz/B_norm, color='red')

        ax.set_xlabel('$x/a_0$')
        ax.set_ylabel('$y/a_0$')
        ax.set_zlabel('$z/a_0$')

        ax.set_xlim([0, self.xmax-self.dx])
        ax.set_ylim([0, self.ymax-self.dy])
        ax.set_zlim([0, self.zmax-self.dz])

        ax.set_aspect('equal')


import plotly.graph_objects as go
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
from comfit.tool.tool_colormap_angle import tool_colormap_angle


def plot_tool_plotly_add_angle_colorbar3D(self, fig: go.Figure) -> go.Figure:
    """Adds a colorbar to a 3D plot with the angle colormap.

    Args:
        fig: The figure to which the colorbar is added.
    
    Returns:
        The figure with the colorbar added.
    """
    

    return fig


from typing import Optional
import numpy as np
import matplotlib

from comfit.plot.plot_complex_field_in_plane_matplotlib import plot_complex_field_in_plane_matplotlib
from comfit.tool.tool_complete_field import tool_complete_field

def plot_angle_field_in_plane_matplotlib(
    self,
    angle_field: np.ndarray,
    normal_vector: Optional[np.ndarray] = None,
    position: Optional[np.ndarray] = None,
    **kwargs
    ) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Plots the angle field in a plane.

    Args:
        angle_field (numpy.ndarray): The angle field to be plotted.
        normal_vector (array-like, optional): The normal vector of the plane. Default is [0,1,0].
        position (array-like, optional): The position of the plane. Default is the middle of the system.
        **kwargs: Keyword arguments for the plot, see https://comfitlib.com/Plotting/.

    Returns:
        tuple: A tuple containing:
            - matplotlib.figure.Figure: The figure containing the plot.
            - matplotlib.axes.Axes: The axes containing the plot.
    """

    # Check if the vector field is complex
    if np.iscomplexobj(angle_field):
        print("\033[91mWarning: the angle vector field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
        print('Max imaginary part: ', np.max(np.imag(angle_field)))
        angle_field = np.real(angle_field)

    # Extend the field if not a complete array is given
    angle_field = tool_complete_field(self, angle_field)

    complex_field = np.exp(1j * angle_field)

    return plot_complex_field_in_plane_matplotlib(self, complex_field, normal_vector=normal_vector, position=position, **kwargs)



import numpy as np
import matplotlib.pyplot as plt
import matplotlib

from comfit.tool.tool_complete_field import tool_complete_field
from comfit.tool.tool_add_spacing_2D import tool_add_spacing_2D
from comfit.tool.tool_add_spacing_3D import tool_add_spacing_3D

from comfit.tool.tool_set_plot_axis_properties_matplotlib import tool_set_plot_axis_properties_matplotlib


def plot_vector_field_matplotlib(self, 
        vector_field, 
        **kwargs) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Plots a vector field.

    Args:
        vector_field (tuple): Tuple containing the x and y components of the vector field.
        **kwargs: 
            spacing (int, optional): The spacing for the quiver plot. Default is self.xRes//20.
            Keyword arguments for the plot, see https://comfitlib.com/Plotting/.

    Returns:
        Tuple containing
            - The figure containing the plot.
            - The axes containing the plot.
    """
    spacing = kwargs.get('spacing', max(self.xRes//20,1))

    # Extend the field if not a complete array is given
    vector_field_copy = []
    for n in range(len(vector_field)):
        vector_field_copy.append(tool_complete_field(self, vector_field[n]))
    vector_field = np.array(vector_field_copy)
    
    # Keyword arguments
    axis_equal = kwargs.get('axis_equal',True)
    colormap = kwargs.get('colormap', 'viridis')

    # Check if the vector field is complex
    if np.iscomplexobj(vector_field):
        print("\033[91mWarning: the provided vector field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
        print('Max imaginary part: ', np.max(np.imag(vector_field)))
        vector_field = np.real(vector_field)


    # Check if an axis object is provided
    fig = kwargs.get('fig', plt.gcf())
    ax = kwargs.get('ax', None)          


    kwargs['plot_is_3D'] = False

    ###############################################################
    ########### DIMENSION: 1 - VECTOR-DIMENSION: 1 ################
    ###############################################################
        
    if self.dim == 1 and vector_field.shape == (1,self.xRes):

        X, Y = np.meshgrid(self.x, np.array([0]), indexing='ij')

        U = np.zeros(X.shape)
        V = np.zeros(X.shape)

        V[:,0] = vector_field[0]

        X,Y,U,V = tool_add_spacing_2D(X,Y,U,V,spacing)


        if ax == None:
            fig.clf()
            ax = plt.gcf().add_subplot(111)

        ax.quiver(X/self.a0, Y/self.a0, U, V, color='blue', angles='xy', scale_units='xy', scale=1)
        kwargs['ylim'] = [np.min(vector_field[0]), np.max(vector_field[0])]
        

    ###############################################################
    ########### DIMENSION: 1 - VECTOR-DIMENSION: 2 ################
    ###############################################################

    elif self.dim == 1 and vector_field.shape == (2,self.xRes):

        kwargs['plot_is_3D'] = True
            
        X, Y, Z = np.meshgrid(self.x, np.array([0]), np.array([0]), indexing='ij')

        U = np.zeros(X.shape)
        V = np.zeros(X.shape)
        W = np.zeros(X.shape)

        V[:,0,0] = vector_field[0]
        W[:,0,0] = vector_field[1]

        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)


        if ax == None:
            fig.clf()
            ax = fig.add_subplot(111, projection='3d')

        ax.quiver(X/self.a0, Y/self.a0, Z/self.a0, U, V, W, color='blue')

        kwargs['ylim'] = [np.min(vector_field[0]), np.max(vector_field[0])]
        delta_y = kwargs['ylim'][1] - kwargs['ylim'][0]

        kwargs['zlim'] = [np.min(vector_field[1]), np.max(vector_field[1])]
        delta_z = kwargs['zlim'][1] - kwargs['zlim'][0]

        if delta_y < 0.15*delta_z:
            kwargs['ylim'] = [kwargs['ylim'][0] - 0.15*delta_z, kwargs['ylim'][1] + 0.15*delta_z]
        
        if delta_z < 0.15*delta_y:
            kwargs['zlim'] = [kwargs['zlim'][0] - 0.15*delta_y, kwargs['zlim'][1] + 0.15*delta_y]

    ###############################################################
    ########### DIMENSION: 1 - VECTOR-DIMENSION: 3 ################
    ###############################################################

    elif self.dim == 1 and vector_field.shape == (3,self.xRes):
        
        kwargs['plot_is_3D'] = True

       
        if ax == None:
            fig.clf()
            ax = fig.add_subplot(111, projection='3d')
        
        X, Y, Z = np.meshgrid(self.x, np.array([0]), np.array([0]), indexing='ij')

        U = np.zeros(X.shape)
        V = np.zeros(X.shape)
        W = np.zeros(X.shape)

        # Add the vector field
        U[:,0,0] = vector_field[0]
        V[:,0,0] = vector_field[1]
        W[:,0,0] = vector_field[2]

        # Add spacing
        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2 + V ** 2 + W ** 2))

        # Normalizing
        U = U/max_vector
        V = V/max_vector
        W = W/max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*self.xmax/15/self.a0)
        vy_scale = kwargs.get('vy_scale', 1)
        vz_scale = kwargs.get('vz_scale', 1)

        # Scaling
        U = vx_scale*U
        V = vy_scale*V
        W = vz_scale*W

        ax.quiver(X/self.a0, Y/self.a0, Z/self.a0, U, V, W, color='blue')

        kwargs['ylim'] = [-1,1]
        kwargs['zlim'] = [-1,1]

    ###############################################################
    ########### DIMENSION: 2 - VECTOR-DIMENSION: 1 ################
    ###############################################################

    elif self.dim == 2 and vector_field.shape == (1,self.xRes,self.yRes):
        
      
        if ax == None:
            fig.clf()
            ax = plt.gcf().add_subplot(111)

        X, Y = np.meshgrid(self.x, self.y, indexing='ij')

        U = vector_field[0]
        V = np.zeros(X.shape)

        X,Y,U,V = tool_add_spacing_2D(X,Y,U,V,spacing)

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2))
        
        U = U / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*self.dx/self.a0)

        # Scaling
        U = vx_scale*U

        ax.quiver(X/self.a0, Y/self.a0, U, V, color='blue',
            angles='xy', scale_units='xy', scale=1,
            headwidth=3, headlength=4, headaxislength=3, pivot='middle')

    ###############################################################
    ########### DIMENSION: 2 - VECTOR-DIMENSION: 2 ################
    ###############################################################

    elif self.dim == 2 and vector_field.shape == (2,self.xRes,self.yRes):

        if ax == None:
            fig.clf()
            ax = plt.gcf().add_subplot(111)

        X, Y = np.meshgrid(self.x, self.y, indexing='ij')

        X, Y, U, V = tool_add_spacing_2D(X,Y,vector_field[0],vector_field[1],spacing)

   

        max_vector = np.max(np.sqrt(U ** 2 + V ** 2))
    
        U = U / max_vector
        V = V / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', 1.0*spacing*self.dx/self.a0)
        vy_scale = kwargs.get('vy_scale', 1.0*spacing*self.dy/self.a0)

        # Scaling
        U = vx_scale*U
        V = vy_scale*V

        ax.quiver(X/self.a0, Y/self.a0, U, V, color='blue', 
            angles='xy', scale_units='xy', scale=1,
            headwidth=3, headlength=4, headaxislength=3, pivot='middle')


    ###############################################################
    ########### DIMENSION: 2 - VECTOR-DIMENSION: 3 ################
    ###############################################################

    elif self.dim == 2 and vector_field.shape == (3,self.xRes,self.yRes):

        kwargs['plot_is_3D'] = True
        
        X, Y, Z = np.meshgrid(self.x, self.y, np.array([0]), indexing='ij')
        U = np.zeros(X.shape)
        V = np.zeros(X.shape)
        W = np.zeros(X.shape)
        
        U[:,:,0] = vector_field[0]
        V[:,:,0] = vector_field[1]
        W[:,:,0] = vector_field[2]

        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        max_vector = np.max(np.sqrt(U ** 2 + V ** 2 + W ** 2))

        # Normalizing
        U = U / max_vector
        V = V / max_vector
        W = W / max_vector
        
        # Scale factors
        vx_scale = kwargs.get('vx_scale', 2*spacing*self.xmax/max_vector/self.a0)
        vy_scale = kwargs.get('vy_scale', 2*spacing*self.ymax/max_vector/self.a0)
        vz_scale = kwargs.get('vz_scale', spacing/self.a0)

        # Scaling
        U = vx_scale*U
        V = vy_scale*V
        W = vz_scale*W

        
        if ax == None:
            fig.clf()
            ax = fig.add_subplot(111, projection='3d')
        ax.quiver(X/self.a0, Y/self.a0, Z/self.a0, U, V, W, color='blue')
       
        kwargs['axis_equal'] = False
        kwargs['zlim'] = [-spacing,spacing]


    ###############################################################
    ########### DIMENSION: 3 - VECTOR-DIMENSION: 1 ################
    ###############################################################

    elif self.dim == 3 and vector_field.shape == (1,self.xRes,self.yRes,self.zRes):

        kwargs['plot_is_3D'] = True

        X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')              

        # Define the vector field
        U = vector_field[0]
        V = np.zeros(U.shape)
        W = np.zeros(U.shape)

        # Add spacing
        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2))
        U = U / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*self.dx/self.a0)

        # Scaling
        U = vx_scale*U


        if ax == None:
            fig.clf()
            ax = fig.add_subplot(111, projection='3d')
        ax.quiver(X/self.a0, Y/self.a0, Z/self.a0, U, V, W, color='blue')
        
    ###############################################################
    ########### DIMENSION: 3 - VECTOR-DIMENSION: 2 ################
    ###############################################################

    elif self.dim == 3 and vector_field.shape == (2,self.xRes,self.yRes,self.zRes):

        kwargs['plot_is_3D'] = True

        X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')

        # Define the vector field
        U = vector_field[0]
        V = vector_field[1]
        W = np.zeros(U.shape)

        # Add spacing
        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2 + V ** 2))
        U = U / max_vector
        V = V / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*self.dx/self.a0)
        vy_scale = kwargs.get('vy_scale', spacing*self.dy/self.a0)

        # Scaling
        U = vx_scale*U
        V = vy_scale*V

      
        if ax == None:
            fig.clf()
            ax = fig.add_subplot(111, projection='3d')

        ax.quiver(X/self.a0, Y/self.a0, Z/self.a0, U, V, W, color='blue')
        
   
    ###############################################################
    ########### DIMENSION: 3 - VECTOR-DIMENSION: 3 ################
    ###############################################################

    elif self.dim == 3 and vector_field.shape == (3,self.xRes,self.yRes,self.zRes):

        kwargs['plot_is_3D'] = True

        X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')

        # Define the vector field
        U = vector_field[0]
        V = vector_field[1]
        W = vector_field[2]

        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)



        if ax == None:
            fig.clf()
            ax = fig.add_subplot(111, projection='3d')

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2 + V ** 2 + W ** 2))

        U = U / max_vector
        V = V / max_vector
        W = W / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*self.dx/self.a0)
        vy_scale = kwargs.get('vy_scale', spacing*self.dy/self.a0)
        vz_scale = kwargs.get('vz_scale', spacing*self.dz/self.a0)

        # Scaling
        U = vx_scale*U
        V = vy_scale*V
        W = vz_scale*W
        ax.quiver(X/self.a0, Y/self.a0, Z/self.a0, U, V, W, color='blue')
        

    ###############################################################
    ###########     NON-VALID DIMENSION            ################
    ###############################################################

    else:
        raise Exception("You have entered an invalid field to the plot_vector_field function.")

    kwargs['ax'] = ax
    tool_set_plot_axis_properties_matplotlib(self, **kwargs)
    return fig, ax





def plot_save_matplotlib(counter, fig, **kwargs):
    """

    Saves the current Matplotlib plot as a PNG file.

    Args:
        fig (matplotlib.figure.Figure, ax): The Matplotlib figure and axis to save as an image.
        counter (int): A unique identifier for the plot image file.

    Keyword Args:
        ID (str, optional): A unique identifier for the plot image file. Defaults to None.
        image_size (tuple, optional): The size of the image in inches. Defaults to (5,5).
        dpi (int, optional): The resolution of the image in dots per inch. Defaults to 100.
        
    
    Returns: 
        None, saves the plot as a PNG file.
    """

    # Keyword arguments
    ID=kwargs.get('ID', None)
    image_size_inches=kwargs.get('image_size_inches', (6,5))
    dpi=kwargs.get('dpi', 100)

    if ID is None:
        filename = f'plot_{counter}.png'
    else:
        filename = f'plot_{counter}_{ID}.png'

    import matplotlib.pyplot as plt

    if isinstance(fig, tuple) and len(fig) == 2 and isinstance(fig[0], plt.Figure) and isinstance(fig[1], plt.Axes):
        fig, ax = fig
    elif isinstance(fig, plt.Figure):
        fig = fig
    else:
        raise ValueError("fig must be a matplotlib Figure or a tuple of (Figure, Axes)")

    fig.set_size_inches(image_size_inches)
    fig.savefig(filename,dpi=dpi)


import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
import matplotlib
from typing import Optional
from comfit.tool.tool_set_plot_axis_properties_matplotlib import tool_set_plot_axis_properties_matplotlib

from comfit.tool.tool_complete_field import tool_complete_field

from skimage.measure import marching_cubes

def plot_vector_field_in_plane_matplotlib(self,
        vector_field: tuple[np.ndarray, np.ndarray],
        position: Optional[np.ndarray] = None,
        normal_vector: Optional[np.ndarray] = None,
        spacing: int = 2,
        **kwargs
    ) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:

    """
    Plots the vector field in a plane.
    
    Args:
        vector_field (tuple): Tuple containing the x and y components of the vector field.
        position (array-like, optional): The position of the plane. Default is the middle of the system.
        normal_vector (array-like, optional): The normal vector of the plane. Default is [0,1,0].
        spacing (int, optional): The spacing for the quiver plot. Default is 5.
        kwargs: Keyword arguments for the plot, see https://comfitlib.com/Plotting/.

    Returns:
        Tuple consisting of
            - The figure containing the plot. matplotlib.figure.Figure.
            - The axes containing the plot. matplotlib.axes.Axes: 
    """

    # Convert the vector field to a numpy array
    vector_field = np.array(vector_field)

    # Extend the field if not a complete array is given
    # TODO: Make this part more efficient (Vidar 11.03.24)
    vector_field_copy = []
    for n in range(vector_field.shape[0]):
        vector_field_copy.append(tool_complete_field(self, vector_field[n]))
    vector_field = np.array(vector_field_copy)

    # Check if the vector field is complex
    if np.iscomplexobj(vector_field):
        print("\033[91mWarning: the provided vector field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
        print('Max imaginary part: ', np.max(np.imag(vector_field)))
        vector_field = np.real(vector_field)
    

    if self.dim != 3:
        raise Exception("The plot in plane function is only defined for 3D fields.")

    # Check if an axis object is provided
    fig = kwargs.get('fig', plt.gcf())
    ax = kwargs.get('ax', None)

    if ax == None:
        ax = fig.add_subplot(111, projection='3d')

    # Default values of position and normal vector
    if position is None:
        position = self.rmid
    
    if normal_vector is None:
        normal_vector = [0,1,0]
    
    normal_vector = np.array(normal_vector)/np.linalg.norm(normal_vector)
    height_above_plane = (self.x-position[0])*normal_vector[0] + (self.y-position[1])*normal_vector[1] + (self.z-position[2])*normal_vector[2]

    verts, faces, _, _ = marching_cubes(height_above_plane, 0)

    # Assuming field is defined on the same grid as height_above_plane
    x, y, z = np.mgrid[0:height_above_plane.shape[0], 0:height_above_plane.shape[1], 0:height_above_plane.shape[2]]

    # Flatten the grid for interpolation
    points = np.c_[x.ravel(), y.ravel(), z.ravel()]

    # Extract the vector field
    if vector_field.shape == (1,self.xRes,self.yRes,self.zRes):
        n = 1
    elif vector_field.shape == (2,self.xRes,self.yRes,self.zRes):
        n = 2
    elif vector_field.shape == (3,self.xRes,self.yRes,self.zRes):
        n = 3
    else:
        raise Exception("You have entered an invalid field to the plot_vector_field function.")

    U = vector_field[0]
    U_values = U.ravel()
    U_verts = sp.interpolate.griddata(points, U_values, verts, method='nearest')

    if n > 1:
        V = vector_field[1]
        V_values = V.ravel()
        V_verts = sp.interpolate.griddata(points, V_values, verts, method='nearest')
    else:
        V_verts = np.zeros(U_verts.shape)
    
    if n > 2:
        W = vector_field[2]
        W_values = W.ravel()
        W_verts = sp.interpolate.griddata(points, W_values, verts, method='nearest')
    else:
        W_verts = np.zeros(U_verts.shape)

    # Normalize the vectors
    max_vector = np.max(np.sqrt(U_verts ** 2 + V_verts ** 2 + W_verts ** 2))
    U_verts = U_verts / max_vector
    V_verts = V_verts / max_vector
    W_verts = W_verts / max_vector

    # Scale factors
    vx_scale = kwargs.get('vx_scale', spacing*self.dx)
    vy_scale = kwargs.get('vy_scale', spacing*self.dy)
    vz_scale = kwargs.get('vz_scale', spacing*self.dz)

    # Scaling
    U_verts = vx_scale*U_verts
    V_verts = vy_scale*V_verts
    W_verts = vz_scale*W_verts

    x = self.xmin+verts[:, 0]*self.dx
    y = self.ymin+verts[:, 1]*self.dy
    z = self.zmin+verts[:, 2]*self.dz

    # Adjust positions based on the coordinate system
    x = self.xmin + x * self.dx
    y = self.ymin + y * self.dy
    z = self.zmin + z * self.dz

    # Add spacing
    x = x[::spacing]
    y = y[::spacing]
    z = z[::spacing]
    U_verts = U_verts[::spacing]
    V_verts = V_verts[::spacing]
    W_verts = W_verts[::spacing]

    ax.quiver(x, y, z, U_verts, V_verts, W_verts, color='blue')

    kwargs['ax'] = ax
    tool_set_plot_axis_properties_matplotlib(self, **kwargs)
    return fig, ax


from typing import Callable, Union, Optional

import numpy as np
import matplotlib.pyplot as plt
from comfit.core.base_system import BaseSystem
from tqdm import tqdm
from mpl_toolkits.mplot3d import Axes3D  
import scipy as sp

class BoseEinsteinCondensate(BaseSystem):
    def __init__(self, dim: int, **kwargs):
        """Initializes a system to simulate a Bose-Einstein Condensate using the Gross-Pitaevskii equation.

        Args:
            dim : int
                The dimension of the system.
            kwargs : dict, optional
                Optional keyword arguments to set additional parameters. see
                https://comfitlib.com/ClassBoseEinsteinCondensate/

        Returns:
            The system object representing the BoseEinsteinCondensate simulation. BoseEinsteinCondensate object

        Example:
            bec = BoseEinsteinCondensate(3,xRes=101,yRes=101,zRes=101, gamma=0.5)
            Creates a BoseEinsteinCondensate system with 3 dimensions and a spatial resolution of 101 in all directions.
            The dissipative factor gamma is set to 0.5.
        """

        # First initialize the BaseSystem class
        super().__init__(dim, **kwargs)

        # Type of the system
        self.psi = None
        self.psi_f = None
        self.type = 'BoseEinsteinCondensate'

        # Default simulation parameters
        self.gamma = 0.01 if 'gamma' not in kwargs else kwargs['gamma']  # Dissipation (gamma)

        self.V_ext = lambda t: 0  # External potential, note this is a function of t

        # If there are additional arguments provided, set them as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self) -> str:
        """Output a string representation of the system.

        Args:
            None

        Returns:
            A string representation of the system.
        """
        return f"ComFiT object: BoseEinsteinCondensate\n \
                Dimension: {self.dim}\n"

    # CONFIGURATION FUNCTIONS
    def conf_initial_condition_disordered(self, noise_strength: float = 0.01) -> np.ndarray:
        """Sets disordered initial condition for the BoseEinsteinCondensate with some thermal flcutiations
        
        Args:
            noise_strength: the strength of the noise

        Returns:
            Sets the value of self.psi and self.psi_f
        """

        if self.dim == 1:
            self.psi = np.random.rand(self.xRes) - 0.5 + \
                       1j * np.random.rand(self.xRes) - 0.5j

        elif self.dim == 2:
            self.psi = np.random.rand(self.xRes, self.yRes) - 0.5 + \
                       1j * np.random.rand(self.xRes, self.yRes) - 0.5j

        elif self.dim == 3:
            self.psi = np.random.rand(self.xRes, self.yRes, self.zRes) - 0.5 + \
                       1j * np.random.rand(self.xRes, self.yRes, self.zRes) - 0.5j
        else:
            raise Exception("Code for this dimension has not yet been implemented.")

        self.psi = noise_strength * self.psi
        self.psi_f = sp.fft.fftn(self.psi)

    def conf_external_potential(self, V_ext: Union[Callable, float], additive: bool = False) -> None:
        """Sets the external potential of the system.

        Args:
            V_ext (function or float): the external potential
            additive (bool, optional): whether to add the new potential to the existing potential or not

        Returns:
            Modifies the value of self.V_ext
        """

        if not callable(V_ext):
            original_V_ext = V_ext  # Preserve the original value of V_ext
            V_ext = lambda t: original_V_ext
        
        if additive:
            self.V_ext = lambda t: self.V_ext(t) + V_ext(t)
        else:
            self.V_ext = V_ext



    def conf_initial_condition_Thomas_Fermi(self) -> None:
        """Finds the Thomas_Fermi ground state.

        Must be precided by an energy relaxation to find the true ground state

        Args:
            None
        
        Returns: 
            Sets the value of self.psi and self.psi_f
        """
        V_0 = np.zeros(self.dims) + self.V_ext(self.time)
        self.psi = np.emath.sqrt(1 - V_0)

        self.psi[V_0 > 1] = 0
        self.psi_f = sp.fft.fftn(self.psi)

    # CONFIGURATION FUNCTIONS
    def conf_insert_vortex(self, charge: int = 1, position: Optional[list[float]] = None):
        """Sets the initial condition for a vortex dipole

        Args:
            charge (int): the charge of the vortex
            position (list): the position of the vortex

        Returns:
            Modifies the value of self.psi and self.psi_f
        """
        if not (self.dim == 2):
            raise Exception("The dimension of the system must be 2 for a single point vortex.")

        if position is None:
            position = [self.xmid, self.ymid]

        if self.psi is None:
            self.conf_initial_condition_Thomas_Fermi()

        self.psi = self.psi * np.exp(1j * self.calc_angle_field_single_vortex(position=position, charge=charge))
        self.psi_f = sp.fft.fftn(self.psi)

    def conf_insert_vortex_dipole(
        self,
        dipole_vector: Optional[list[float]] = None,
        dipole_position: Optional[list[float]] = None
    ):
        """Sets the initial condition for a vortex dipole configuration in a 2-dimensional system.

        Args:
            None

        Returns:
            Modifies the value of self.psi and self.psi_f

        Raises:
            Exception: If the dimension of the system is not 2.
        """
        if not (self.dim == 2):
            raise Exception("The dimension of the system must be 2 for a vortex dipole configuration.")

        if self.psi is None:
            self.conf_initial_condition_Thomas_Fermi()

        if dipole_vector is None:
            dipole_vector = [(self.xmax-self.xmin) / 3, 0]

        if dipole_position is None:
            dipole_position = self.rmid

        self.psi = self.psi * np.exp(1j * self.calc_angle_field_vortex_dipole(dipole_vector, dipole_position))
        self.psi_f = sp.fft.fftn(self.psi)

    def conf_insert_vortex_ring(
        self,
        position: Optional[list[float]] = None,
        radius: Optional[float] = None,
        normal_vector: list[float] = [0, 0, 1]
    ) -> None:
        """Sets the initial condition for a vortex ring configuration in a 3-dimensional system
        
        Args:
            position: the position of the vortex ring (list)
            radius: the radius of the vortex ring (float)
            normal_vector: the normal vector of the vortex ring (list)
        
        Returns:
            Modifies the value of self.psi and self.psi_f
        """
        if not (self.dim == 3):
            raise Exception("The dimension of the system must be 3 for a vortex ring configuration.")

        if position is None:
            position = self.rmid

        if radius is None:
            radius = (self.xmax -self.xmin)/ 3

        theta = self.calc_angle_field_vortex_ring(position=position, radius=radius, normal_vector=normal_vector)

        if self.psi is None:
            self.conf_initial_condition_Thomas_Fermi()

        self.psi = self.psi * np.exp(1j * theta)
        self.psi_f = sp.fft.fftn(self.psi)

    def conf_vortex_remover(self, nodes: list[dict], Area: list[float]) -> None:
        '''Removes vortices
        
        Function that finds and removes vortices outside of the area defined by the corners
        (x1,y1), (x1,y2), (x2,y1), (x2,y2)

        Args:
            nodes (list) a list containing the vortices
            Area  (array) list on the format (x1,x2,y1,y2)

        Returns:
            Modifies the value of self.psi and self.psi_f
        '''
        for vortex in nodes:
            x_coord = vortex['position'][0]
            y_coord = vortex['position'][1]
            if not (Area[0] < x_coord and x_coord < Area[1] \
                    and Area[2] < y_coord and y_coord < Area[3]):
                self.conf_insert_vortex(charge=-1 * vortex['charge'], position=[x_coord + self.dx, y_coord])
                # self.conf_insert_vortex(charge=vortex['charge'], position=[7, 0])

    def conf_dissipative_frame(self, interface_width: float = 7, 
                                    frame_width_x: float = None, 
                                    frame_width_y: float = None, 
                                    frame_width_z: float = None) -> None:

        '''Configures a dissipative frame around the computational domain

        This function sets self.gamma so that it has a low value in the bulk and a large value near the edges.
        This sets a dissipative frame around the computational domain
        
        Args:
            d: length of the interface between the low gamma and high gamma regions (float)
            frame_width_x: distance fom center to the frame in x-direction (float)
            frame_width_y:    -- " --                         y-direction (float)
            frame_width_z:     -- " --                         z-direction (float)

        Returns:
            modify self.gamma
        '''

        # If configuration is called for the first time, set initial_gamma
        if not hasattr(self, 'initial_gamma'):
            self.initial_gamma = self.gamma

        # Set default values for frame_width
        if frame_width_x is None:
            frame_width_x = 0.8*self.size_x
        if frame_width_y is None:
            frame_width_y = 0.8*self.size_y
        if frame_width_z is None:
            frame_width_z = 0.8*self.size_z

        # Set the frame
        if self.dim == 2:
            X, Y = np.meshgrid(self.x, self.y, indexing='ij')
            gammax = self.initial_gamma + 0.5 * (2 + np.tanh((X - self.xmid - frame_width_x/2) /interface_width) - np.tanh((X - self.xmid + frame_width_x/2) /interface_width))
            gammay = self.initial_gamma + 0.5 * (2 + np.tanh((Y - self.ymid - frame_width_y/2) /interface_width) - np.tanh((Y - self.ymid + frame_width_y/2) /interface_width))
            self.gamma = np.real(np.maximum(gammax, gammay))
        elif self.dim == 3:
            X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')
            gammax = self.initial_gamma + 0.5 * (2 + np.tanh((X - self.xmid - frame_width_x/2) /interface_width) - np.tanh((X - self.xmid + frame_width_x/2) /interface_width))
            gammay = self.initial_gamma + 0.5 * (2 + np.tanh((Y - self.ymid - frame_width_y/2) /interface_width) - np.tanh((Y - self.ymid + frame_width_y/2) /interface_width))
            gammaz = self.initial_gamma + 0.5 * (2 + np.tanh((Z - self.zmid - frame_width_z/2) /interface_width) - np.tanh((Z - self.zmid + frame_width_z/2) /interface_width))
            self.gamma = np.real(np.maximum(gammax, gammay, gammaz))
        else:
            raise Exception("This feature is not yet available for the given dimension.")


    # Time evolution
    def evolve_dGPE(self, number_of_steps: int , method: str = 'ETD2RK') -> None:
        '''Evolver for the dGPE.
        
        Args:
            number_of_steps: the number of time steps that we are evolving the equation  (int)
            method: the integration method we want to use. ETD2RK is sett as default  (string, optional)
        
        Returns:
            Updates the self.psi and self.psi_f
       '''

        k2 = self.calc_k2()
        omega_f = (1j + self.gamma) * (1 - 1 / 2 * k2)

        integrating_factors_f, solver = self.calc_integrating_factors_f_and_solver(omega_f, method)

        for n in tqdm(range(number_of_steps), desc='evolving the dGPE'):
            self.psi, self.psi_f = solver(integrating_factors_f,
                                          self.calc_nonlinear_evolution_function_f,
                                          self.psi, self.psi_f)

    def evolve_relax(self, number_of_steps: int, method: str = 'ETD2RK') -> None:
        '''Evolver for the dGPE in imaginary time that relax the equation closer to the ground state
        
        Args:
            number_of_steps: the number of time steps that we are evolving the equation (int)
            method: the integration method we want to use. ETD2RK is sett as default (string)
        
        Returns:
            Updates the self.psi and self.psi_f
        '''
        temp_t = self.time
        gamma0 = self.gamma

        self.gamma = 1 - 1j

        temp_V = self.V_ext

        self.conf_external_potential(temp_V(temp_t))

        print("Relaxing the BoseEinsteinCondensate...")
        self.evolve_dGPE(number_of_steps, method)

        self.gamma = gamma0
        self.time = temp_t
        
        self.conf_external_potential(temp_V)

    def evolve_comoving_dGPE(self, number_of_steps: int, velx: float, method: str = 'ETD2RK') -> None:
        '''Evolver for the dGPE in the comoving frame.

        This evolver assume that the stirring is in the x-direction and that gamma is spatialy dependent
        
        Args:
            number_of_steps: the number of time steps that we are evolving the equation (int)
            velx: velocity in x direction (float) 
            method: the integration method we want to use. ETD2RK is sett as default (string)
        
        Returns:
            Updates the fields self.psi and self.psi_f
        '''
        k2 = self.calc_k2()

        omega_f = (1j) * (1 - 1 / 2 * k2) + velx * self.dif[0]

        if method == 'ETD2RK':
            integrating_factors_f = self.calc_evolution_integrating_factors_ETD2RK(omega_f)
            solver = self.evolve_ETD2RK_loop
        elif method == 'ETD4RK':
            integrating_factors_f = self.calc_evolution_integrating_factors_ETD4RK(omega_f)
            solver = self.evolve_ETD4RK_loop
        else:
            raise Exception('This method is not implemented')

        for n in tqdm(range(number_of_steps), desc='evolving the dGPE in the comoving frame'):
            self.psi, self.psi_f = solver(integrating_factors_f, self.calc_nonlinear_evolution_term_comoving_f,
                                          self.psi, self.psi_f)

    # CALCULATION FUNCTIONS

    def calc_nonlinear_evolution_function_f(self, psi: np.ndarray, t) -> np.ndarray:
        """Calculates the non-linear evolution term of the dGPE
        
        Args:
            psi: the wavefunction at a given time. (numpy.ndarray)
        
        Returns:
            The non-linear evolution term (numpy.ndarray)
        """
        
        psi2 = np.abs(psi) ** 2
        return sp.fft.fftn((1j + self.gamma) * (-self.V_ext(t) - psi2) * psi)

    def calc_nonlinear_evolution_term_comoving_f(self, psi: np.ndarray, t) -> np.ndarray:
        """Calculates the non-linear evolution term of the dGPE when gamma is not a constant.

        Relevant for example in the comoving frame when we have a dissipative frame around the edge.
        
        Args:
            psi: the wavefunction at a given time. (numpy.ndarray)
        
        Returns:
            the non-linear evolution term (numpy.ndarray)
        """
        psi2 = np.abs(psi) ** 2
        term1 = sp.fft.fftn(-(1j + self.gamma) * (self.V_ext(t) + psi2) * psi)
        term2 = sp.fft.fftn(self.gamma * psi)
        term3 = 0.5 * sp.fft.fftn(self.gamma * sp.fft.ifftn(-self.calc_k2() * sp.fft.fftn(psi)))
        return (term1 + term2 + term3)

        # Functions for callculationg properties of the BoseEinsteinCondensate

    def calc_superfluid_current(self) -> np.ndarray:
        """Calculates the superfluid current

        Args:
            None
        
        Returns:
            The superfluid current (numpy.ndarray) 
        """
        if self.dim == 2:
            J_s = np.zeros((self.dim,self.xRes,self.yRes))
        elif self.dim == 3:
            J_s = np.zeros((self.dim, self.xRes, self.yRes,self.zRes))
        else:
            raise(Exception('Calculation of the  superfluid current is not implemented in this dimension'))
        for i in range(self.dim):
            J_s[i] = np.imag( np.conj(self.psi) * sp.fft.ifftn(1j*self.k[i] *self.psi_f ))
        return J_s

    def calc_velocity(self) -> np.ndarray:
        """Calculates the weighted velocity field
        
        Args:
            None

        Returns:
            The weighted velocity field (numpy.ndarray) 
        """
        if self.dim == 2:
            u = np.zeros((self.dim, self.xRes, self.yRes))
        elif self.dim == 3:
            u = np.zeros((self.dim, self.xRes, self.yRes, self.zRes))
        else:
            raise (Exception('Calculation of the weighted velocity is not implemented in this dimension'))
        theta = np.angle(self.psi)
        for i in range(self.dim):
            u[i] = np.imag(np.exp(-1j*theta)* sp.fft.ifftn(1j * self.k[i] * self.psi_f))
        return u

    def calc_kinetic_energy(self) -> float:
        """Calculates the kinetic energy.

        Args:
            None

        Returns:
            The kinetic energy (float) 
        """
        u = self.calc_velocity()
        u2 = np.sum(u[i]**2 for i in range(self.dim))
        return 0.5*self.calc_integrate_field(u2)


    def calc_hamiltonian_density(self) -> np.ndarray:
        """Calculates the hamiltonian density

        Args:
            None

        Returns:
            The hamiltonian density (numpy.ndarray) 
        """
        k2 = self.calc_k2()
        interaction_term = 1/2*np.abs(self.psi)**4
        potential_term = (self.V_ext(self.time) - 1 )* np.abs(self.psi)**2
        laplacian_term = -1/2 *np.real( np.conj(self.psi) * sp.fft.ifftn(-k2*self.psi_f))
        return laplacian_term +potential_term + interaction_term

    def calc_hamiltonian(self) -> float:
        """Function that calculates the Hamiltonian

        Args:
            None

        Returns:
            The Hamiltonian
        """
        H = self.calc_hamiltonian_density()
        return self.calc_integrate_field(H)

    def calc_harmonic_potential(self, R_tf: float) -> np.ndarray:
        """Calculates a harmonic trap with R_tf being the Thomas-Fermi radius
        
        Args:
            R_tf (float): The Thomas-Fermi radius
        
        Returns:
            A harmonic potential
        """
        trapping_strength = 1 / (R_tf ** 2)
        if self.dim == 1:
            return trapping_strength * (self.x - self.xmid) ** 2
        if self.dim == 2:
            return trapping_strength * (((self.x - self.xmid) ** 2).reshape(self.xRes, 1)
                                        + ((self.y - self.ymid) ** 2).reshape(1, self.yRes))
        if self.dim == 3:
            return trapping_strength * (((self.x - self.xmid) ** 2).reshape(self.xRes, 1, 1)
                                        + ((self.y - self.ymid) ** 2).reshape(1, self.yRes, 1)
                                        + ((self.z - self.zmid) ** 2).reshape(1, 1, self.zRes))

    def calc_gaussian_stirring_potential(self, size: float, strength: float, position: np.ndarray) -> np.ndarray:
        """Calculates a gaussian potential
        
        Args:
            size (float) size of the stirrer
            strength (float) strength of the stirrer, i.e how large the potential is at the stirrers position
            position (array) the position of the stirrer
       
        Retuns:
            A gaussian potential (numpy.ndarray) 
        """

        if self.dim == 1:
            return strength * np.exp(-(self.x - position[0]) ** 2 / size ** 2)

        elif self.dim == 2:
            return strength * np.exp(-(((self.x - position[0]) ** 2).reshape(self.xRes, 1)
                                       + ((self.y - position[1]) ** 2).reshape(1, self.yRes)) / (size ** 2))
        elif self.dim == 3:
            return strength * np.exp(-(((self.x - position[0]) ** 2).reshape(self.xRes, 1, 1)
                                       + ((self.y - position[1]) ** 2).reshape(1, self.yRes, 1)
                                       + ((self.z - position[2]) ** 2).reshape(1, 1, self.zRes)) / (size ** 2))

    def calc_force_on_external_potential(self) -> np.ndarray: 
        """Calculates the average force acting on the external potential.
        
        Args:
            None
        
        Returns:
            Average force on the potential (numpy.ndarray) 
        """
        Force =np.zeros(self.dim)
        potential_f = sp.fft.ifftn(self.V_ext(self.time))
        for i in range(self.dim):
            Force_density = np.real(np.abs(self.psi)**2 * sp.fft.ifftn(1j*self.k[i]* potential_f))
            Force[i] = self.calc_integrate_field(Force_density)
        return Force
        #TODO: It is not clear to me exactly what this function does (Vidar 04.12.23)



    ## Functions for calculating vortex properties
    def calc_vortex_density(self, psi: Optional[np.ndarray] = None) -> np.ndarray:
        """Calculates the vortex density of the system.

        Args:
            psi: The wavefunction of the system. (numpy.ndarray)
        
        Returns:
            The vortex density of the system. numpy.ndarray
        """
        if psi is None:
            psi = self.psi

        return self.calc_defect_density([np.real(psi), np.imag(psi)])

    def calc_vortex_density_singular(self) -> np.ndarray:
        """Calculates the vortex density of the system using the singular method.

        Args:
            None
        
        Returns:
            The vortex density of the system. (numpy.ndarray)
        """
        # TODO: Insert the correct value of the equilibrium of psi, based on theory (Vidar 03.12.23)
        return self.calc_defect_density([np.real(self.psi), np.imag(self.psi)])

    def calc_vortex_velocity_field(self, dt_psi: np.ndarray, psi: Optional[np.ndarray] = None) -> np.ndarray:
        """Calculates the vortex velocity field of the system.

        Args:
            dt_psi: The time derivative of the wavefunction of the system. (numpy.ndarray)
            psi: The wavefunction of the system. (numpy.ndarray)

        Returns:
            The vortex velocity field of the system. numpy.ndarray: 
        """
        if psi is None:
            psi = self.psi

        return self.calc_defect_velocity_field([np.real(psi), np.imag(psi)],
                                        [np.real(dt_psi), np.imag(dt_psi)])

    def calc_vortex_nodes(self, dt_psi: Optional[np.ndarray] = None) -> list[dict]:
        """
        Calculate the positions and charges of vortex nodes based on the defect density.

        Args:
            dt_psi: The time derivative of the wavefunction of the system.     (numpy.ndarray)
    
        Returns:
            List of dictionaries representing the vortex nodes. Each dictionary contains the following keys:
                  - 'position_index': The position index of the vortex node in the defect density array.
                  - 'charge': The charge of the vortex node.
                  - 'position': The position of the vortex node as a list [x, y].
                  - 'velocity': The velocity of the vortex node as a list [vx, vy].
        """

        # Calculate defect density
        rho = self.calc_vortex_density(self.psi)

        if dt_psi is not None:
            velocity_field = self.calc_vortex_velocity_field(dt_psi, self.psi)

        if self.dim == 2:
            vortex_nodes = self.calc_defect_nodes(np.abs(rho), 
                                                    charge_tolerance = 0.2,
                                                    integration_radius = self.a0)
            for vortex in vortex_nodes:
                vortex['charge'] = np.sign(rho[vortex['position_index']])
                if dt_psi is not None:
                    vortex['velocity'] = [velocity_field[0][vortex['position_index']], 
                                        velocity_field[1][vortex['position_index']]]
                else:
                    vortex['velocity'] = [float('nan'), float('nan')]
        elif self.dim == 3:
            vortex_nodes = self.calc_defect_nodes(np.sqrt(rho[0]**2 + rho[1]**2 + rho[2]**2), 
                                                charge_tolerance = 2*self.a0,
                                                integration_radius = 2*self.a0
                )
            for vortex in vortex_nodes:
                tangent_vector = np.array([rho[i][vortex['position_index']] for i in range(3)]),
                vortex['tangent_vector'] = tangent_vector[0]/np.linalg.norm(tangent_vector)
                
                if dt_psi is not None:
                    vortex['velocity'] = [velocity_field[0][vortex['position_index']], 
                                        velocity_field[1][vortex['position_index']], 
                                        velocity_field[2][vortex['position_index']]]
                else:
                    vortex['velocity'] = [float('nan'), float('nan'), float('nan')]

        return vortex_nodes



    





def tool_print_in_color(string, color='red'):
    """Print a string in color.

    Parameters
    ----------
    string : str
        The string to be printed.
    color : str
        The color of the string. Options are 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', 'black'.

    Returns
    -------
    None
    """
    colors = {'red': 31, 'green': 32, 'yellow': 33, 'blue': 34, 'magenta': 35, 'cyan': 36, 'white': 37, 'black': 30}
    print('\033[1;{}m{}\033[0m'.format(colors[color], string))
    return None


import numpy as np
import matplotlib.pyplot as plt
import plotly.graph_objects as go

from typing import Union

from comfit.tool import tool_complete_field
from comfit.tool import tool_set_plot_axis_properties_matplotlib
from comfit.tool import tool_set_plot_axis_properties_plotly

from comfit.tool import tool_colormap_bluewhitered
from comfit.tool import tool_colormap_sunburst
from comfit.plot import plot_surface_matplotlib

import matplotlib

def plot_field_matplotlib(self, 
        field: np.ndarray, 
        **kwargs
        ) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Plots the given (real) field.
    
    Args:
        field (array-like): The field to be plotted.
        **kwargs: Keyword arguments for the plot.
            See https://comfitlib.com/ClassBaseSystem/ 
            for a full list of keyword arguments.
    
    Returns:
        tuple: A tuple containing:
            - matplotlib.figure.Figure: The figure containing the plot.
            - matplotlib.axes.Axes: The axes containing the plot
    """

    if field.dtype == bool:
        field = field.astype(float)

    if field.ndim == self.dim+1:
        print("\033[91mWarning - in plot_field: the provided field seems to be an order parameter containing several fields. Only the zeroth component will be plotted.\033[0m")
        field = field[0]

    # Check if the vector field is complex
    if np.iscomplexobj(field):
        print("\033[91mWarning - in plot_field: the provided field was complex. This might be due to residual imaginary parts from the Fourier transform. The imaginary parts will be removed.\033[0m")
        print('Max imaginary part: ', np.max(np.imag(field)))
        field = np.real(field)

    # Check if an axis object is provided
    fig = kwargs.get('fig', plt.gcf())
    ax = kwargs.get('ax', None)

    # Kewyord arguments
    colorbar = kwargs.get('colorbar', True)
    axis_equal = kwargs.get('axis_equal', None)

    # Extend the field if not a complete array is given
    field = tool_complete_field(self, field)

    kwargs['plot_is_3D'] = False

    ###############################################################
    ###################### DIMENSION: 1 ###########################
    ###############################################################

    if self.dim == 1:

        # Keyword arguments particular to the 1D case
        kwargs['grid'] = kwargs.get('grid', True)

        if axis_equal is None:
            kwargs['axis_equal'] = False

        if ax == None:
            fig.clf()
            ax = fig.add_subplot(111)

        ax.plot(self.x/self.a0, field)

    ###############################################################
    ###################### DIMENSION: 2 ###########################
    ###############################################################

    if self.dim == 2:
        
        # Keyword arguments particular to the 2D case
        kwargs['grid'] = kwargs.get('grid', False)


        if ax == None:
            fig.clf()
            ax = fig.add_subplot(111)
            
        # Set the colormap
        colormap = kwargs.get('colormap', 'viridis')

        if colormap == 'bluewhitered':
            colormap = tool_colormap_bluewhitered()

        elif colormap == 'sunburst':
            colormap = tool_colormap_sunburst()
        else:
            colormap = plt.get_cmap(colormap)

        # Value limits symmetric
        vlim_symmetric = kwargs.get('vlim_symmetric', False)

        X = kwargs.get('X', None)
        Y = kwargs.get('Y', None)

        if X is None or Y is None:
            X, Y = np.meshgrid(self.x, self.y, indexing='ij')

        pcm = ax.pcolormesh(X / self.a0, Y / self.a0, field, shading='gouraud', cmap=colormap)

        xlim = [self.xmin, self.xmax-self.dx]
        ylim = [self.ymin, self.ymax-self.dy]

        limits_provided = False
        if 'xlim' in kwargs:
            xlim = kwargs['xlim']
            limits_provided = True
        else:
            if 'xmin' in kwargs:
                xlim[0] = kwargs['xmin']
                limits_provided = True
            
            if 'xmax' in kwargs:
                xlim[1] = kwargs['xmax']
                limits_provided = True

        if 'ylim' in kwargs:
            ylim = kwargs['ylim']
            limits_provided = True
        else:
            if 'ymin' in kwargs:
                ylim[0] = kwargs['ymin']
                limits_provided = True
                
            if 'ymax' in kwargs:
                ylim[1] = kwargs['ymax']
                limits_provided = True

        # If explicit limits are provided, use them to change the vlim ranges
        if limits_provided:
            region_to_plot = np.zeros(self.dims).astype(bool)
            region_to_plot[(xlim[0] <= X)*(X <= xlim[1])*(ylim[0] <= Y)*(Y <= ylim[1])] = True
            vlim = [np.min(field[region_to_plot]), np.max(field[region_to_plot])]

        else:
            vlim = [np.min(field), np.max(field)]
        
        # Set the value limits
        if 'vlim' in kwargs:
            vlim = kwargs['vlim']
        else:
            if 'vmin' in kwargs:
                vlim[0] = kwargs['vmin']
            if 'vmax' in kwargs:
                vlim[1] = kwargs['vmax']

        if vlim[1] - vlim[0] < 1e-10:
            vlim = [vlim[0]-0.05, vlim[1]+0.05]

        pcm.set_clim(vmin=vlim[0], vmax=vlim[1])

        if 'vlim_symmetric' in kwargs:
            vlim_symmetric = kwargs['vlim_symmetric']
            if vlim_symmetric:
                cmax = abs(field).max()
                cmin = -cmax
                pcm.set_clim(vmin=cmin, vmax=cmax)

        colorbar = kwargs.get('colorbar', True)

        if colorbar:
            cbar = plt.colorbar(pcm, ax=ax)

    ###############################################################
    ###################### DIMENSION: 3 ###########################
    ###############################################################

    elif self.dim == 3:

        kwargs['plot_is_3D'] = True

        # Keyword arguments particular to the 3D case

        field_min = np.min(field)
        field_max = np.max(field)

        number_of_layers = kwargs.get('number_of_layers', 1)
        alpha = kwargs.get('alpha', 0.5)
        
        if 'vlim' in kwargs:
            vlim = kwargs['vlim']
            vmin = vlim[0]
            vmax = vlim[1]
        else:
            vmin = field_min
            vmax = field_max

        if 'layer_values' in kwargs:
            layer_values = np.concatenate([[-np.inf], kwargs['layer_values'], [np.inf]])
        else: 
            layer_values = np.linspace(vmin, vmax, number_of_layers + 2)

        if 'colormap' in kwargs:
            colormap = kwargs['colormap']
            if colormap == 'bluewhitered':
                colormap = tool_colormap_bluewhitered()

            elif colormap == 'sunburst':
                colormap = tool_colormap_sunburst()

            else:
                colormap = plt.get_cmap(colormap)
        else: 
            colormap = plt.get_cmap('viridis')
        

        if 'colorbar' in kwargs:
            colorbar = kwargs['colorbar']
        else:
            colorbar = True

        #Plotting the layers

        if ax == None:
            plt.clf()
            ax = plt.gcf().add_subplot(111, projection='3d')


        if field_min < layer_values[1] < field_max:
            plot_surface_matplotlib(self, field=field, 
                                    value=layer_values[1], 
                                    color=colormap((layer_values[1]-vmin) / (vmax-vmin)), 
                                    alpha=alpha,
                                    ax=ax)

        for layer_value in layer_values[2:-1]:
            if field_min < layer_value < field_max:
                plot_surface_matplotlib(self, field=field, 
                                        value=layer_value, 
                                        color=colormap((layer_value-vmin) / (vmax-vmin)), 
                                        alpha=alpha,
                                        ax=ax)

        if colorbar:
            sm = plt.cm.ScalarMappable(cmap=colormap)
            sm.set_clim(vmin, vmax)
            plt.colorbar(sm, ax=ax, pad=0.2)

    kwargs['ax'] = ax
    ax = tool_set_plot_axis_properties_matplotlib(self, **kwargs)
    return fig, ax



import numpy as np
from scipy.fftpack import fftn, ifftn
from scipy.special import factorial
from collections import Counter

def tool_multinom(*args):
    M = len(args)
    if M < 2:
        return 1
    else:
        indices = np.array(args)
        counts = Counter(indices)
        prod = np.prod([factorial(count) for count in counts.values()])
        return 1 / prod

def levi_civita_symbol(i, j, k):
    if {i, j, k} != {0, 1, 2}:
        return 0  # Levi-Civita symbol is zero if indices are not 0, 1, 2

    if (i, j, k) in [(0, 1, 2), (1, 2, 0), (2, 0, 1)]:
        return 1  # Even permutation
    elif (i, j, k) in [(2, 1, 0), (0, 2, 1), (1, 0, 2)]:
        return -1  # Odd permutation
    else:
        return 0  # Repeated index

if __name__ == "__main__":
    print(tool_multinom(1,2,2))


"""
This module provides tool for exporting plots.
"""
import matplotlib.pyplot as plt
import comfit as cf

from tqdm import tqdm

def tool_export_rotating_plot(ax=None,step=5):
    if ax is None:
        ax = plt.gca()

    for n in tqdm(range(0,360//step),desc="Saving img for animation"):

        ax.view_init(elev=30, azim=-60+n*step)
        cf.tool_save_plot_matplotlib(n)

    cf.tool_make_animation_gif(n,name="rotating_plot",fps=12)